# -*- coding: utf-8 -*-
import weakref
import time
import zmq
import logging
import queue
import threading
import _thread as thread
import random
from abc import ABCMeta, abstractmethod

from .base import *

__ALL__ = (
    'ZMQIPSub',
)

FRAME_HEARTBEAT = "__HeyDude__"

BROKER_PUB_HWM = 1000
BROKER_REP_HWM = 1000
LISTENER_REQ_HWM = 1
LISTENER_SUB_HWM = 1000

INPROC_HWM = BROKER_PUB_HWM * 2

MIN_UPDATE_REPLY_FRAMES = 1
MAX_UPDATE_REPLY_FRAMES = 1
MAX_UPDATE_REPLY_FIRSTFRAME = 10

if hasattr(zmq, 'HWM'):
    # Has single HWM
    def set_hwm(sock, hwm):
        sock.hwm = hwm
else:
    # Split HWM
    def set_hwm(sock, hwm):  # lint:ok
        sock.set_hwm(hwm)

class BootstrapNow(Exception):
    pass

class ZMQIPSub(BaseIPSub):
    class FSM(object):
        class State(object):
            __metaclass__ = ABCMeta

            def __init__(self, owner, logger = None):
                self._owner = weakref.ref(owner)
                self.logger = getattr(owner, 'logger', logging.getLogger('chorde.ipsub')) if logger is None else logger

            def transition(self, newstate):
                self.logger.debug("ZMQIPSub.FSM: LEAVE %s", self.__class__.__name__)
                self.leave()
                self.__class__ = newstate
                self.logger.debug("ZMQIPSub.FSM: ENTER %s", self.__class__.__name__)
                self.enter()

            @abstractmethod
            def enter(self):
                pass

            @abstractmethod
            def leave(self):
                pass

            @abstractmethod
            def stay(self):
                pass

        class Bootstrap(State):
            def enter(self):
                owner = self._owner()

                # Initialize pull socket so others can connect rightaway
                owner._pullsocket()

                # Try a few times
                for i in range(5):
                    try:
                        owner._bind()
                        break
                    except zmq.ZMQError as e:
                        if e.errno in (zmq.EADDRINUSE, zmq.ENODEV):
                            # Not a transient error, shortcut to listener
                            return self.transition(ZMQIPSub.FSM.Listener)
                    except Exception as e:
                        self.logger.info("Got %s connecting", e)
                        self.logger.debug("Got %s connecting", e, exc_info = True)
                        time.sleep(0.2)
                else:
                    try:
                        owner._bind()
                    except:
                        return self.transition(ZMQIPSub.FSM.Listener)
                self.transition(ZMQIPSub.FSM.DesignatedBroker)

            def stay(self):
                pass

            def leave(self):
                pass

        class Listener(State):
            def enter(self):
                owner = self._owner()
                self.listener_req = listener_req = owner._connect()
                self.listener_sub = listener_sub = owner._subscribe()
                self.pull = pull = owner._pullsocket()
                self.poller = poller = zmq.Poller()
                poller.register(listener_sub, zmq.POLLIN)
                poller.register(pull, zmq.POLLIN)
                poller.register(listener_req, zmq.POLLIN | zmq.POLLOUT)
                owner._notify_all(EVENT_ENTER_LISTENER, None)

            def stay(self):
                owner = self._owner()

                # Cache globals and attributes, to avoid memory allocations in the hottest loop of all
                listener_req = self.listener_req
                listener_sub = self.listener_sub
                pull = self.pull
                poller = self.poller
                POLLOUT = zmq.POLLOUT
                POLLIN = zmq.POLLIN
                HEARTBEAT_ = FRAME_HEARTBEAT
                poller_poll = poller.poll
                poller_register = poller.register
                pull_recv_multipart = pull.recv_multipart
                no_updates = owner.updates.empty
                put_nowait = owner.updates.put_nowait
                send_update = owner._send_update
                recv_update = owner._recv_update
                recv_update_reply = owner._recv_update_reply
                int_ = int
                len_ = len
                dict_ = dict
                random_ = random.random
                tic_count = 100
                hb_period_base = owner.heartbeat_avg_period * 2
                hb_period_spread = hb_period_base * 2
                hb_timeout = owner.heartbeat_push_timeout

                # Poll sockets
                try:
                    while not owner.stop:
                        if owner._needs_subscriptions:
                            owner._subscribe()
                        if not no_updates():
                            poller_register(listener_req, POLLIN|POLLOUT)
                        activity = poller_poll(hb_period_base + int_(hb_period_spread * random_()))
                        if not activity:
                            # Heartbeat gap
                            # Try to send a heartbeat
                            if not listener_req.poll(hb_timeout, POLLOUT):
                                # Must be mute... dead broker?
                                self.logger.warn("Mute req socket: dead broker? bootstrapping")
                                raise BootstrapNow
                            else:
                                listener_req.send(HEARTBEAT_)
                                if not listener_req.poll(hb_timeout):
                                    # Dead again
                                    self.logger.warn("No reply to heartbeat: dead broker? bootstrapping")
                                    raise BootstrapNow
                                else:
                                    # Alive... lets validate the heartbeat pong and move on
                                    if listener_req.recv() != HEARTBEAT_:
                                        # Bad bad bad
                                        self.logger.error("IPSub: bad heartbeat")
                            owner._idle()
                            break
                        tic_count -= 1
                        if tic_count < 0:
                            owner._tic()
                            tic_count = 100
                        # First send all sendable, empty the queue before trying to fill it
                        activity = dict_(activity)
                        if listener_req in activity:
                            what = activity.pop(listener_req)
                            if what & POLLOUT:
                                if not no_updates():
                                    send_update(listener_req)
                                else:
                                    poller_register(listener_req, POLLIN)
                            if what & POLLIN:
                                recv_update_reply(listener_req)
                        # Then put incoming stuff on the queue
                        for socket, what in activity.items():
                            if socket is pull:
                                pack = pull_recv_multipart()
                                if len_(pack) > 1:
                                    # ^ else Wakeup call, ignore
                                    put_nowait(pack)
                                elif pack[0] == "tic":
                                    owner._tic()
                                    tic_count = 100
                                del pack
                            elif socket is listener_sub and what & POLLIN:
                                recv_update(listener_sub)
                except queue.Full:
                    self.logger.error("While handling IPSub FSM pipe: Queue full, update lost")
                except BootstrapNow:
                    self.transition(ZMQIPSub.FSM.Bootstrap)
                except:
                    self.logger.error("Exception in IPSub listener, re-bootstrapping in a sec", exc_info = True)
                    time.sleep(1)
                    self.transition(ZMQIPSub.FSM.Bootstrap)

            def leave(self):
                owner = self._owner()
                if owner is not None:
                    owner._notify_all(EVENT_LEAVE_LISTENER, None)
                    owner._disconnect()

        class DesignatedBroker(State):
            def enter(self):
                owner = self._owner()
                self.broker_rep = broker_rep = owner._repsocket()
                self.broker_pub = broker_pub = owner._pubsocket()
                self.pull = pull = owner._pullsocket()
                self.poller = poller = zmq.Poller()
                poller.register(broker_pub, zmq.POLLOUT)
                poller.register(broker_rep, zmq.POLLIN)
                poller.register(pull, zmq.POLLIN)
                owner._notify_all(EVENT_ENTER_BROKER, None)

            def stay(self):
                owner = self._owner()

                # Cache globals and attributes, to avoid memory allocations in the hottest loop of all
                broker_rep = self.broker_rep
                broker_pub = self.broker_pub
                pull = self.pull
                poller = self.poller
                POLLOUT = zmq.POLLOUT
                POLLIN = zmq.POLLIN
                HEARTBEAT_ = FRAME_HEARTBEAT
                T = True
                poller_poll = poller.poll
                poller_register = poller.register
                poller_unregister = poller.unregister
                pull_recv_multipart = pull.recv_multipart
                no_updates = owner.updates.empty
                put_nowait = owner.updates.put_nowait
                send_update = owner._send_update
                int_ = int
                len_ = len
                dict_ = dict
                random_ = random.random
                tic_count = 100
                hb_period_base = owner.heartbeat_avg_period // 2
                hb_period_spread = hb_period_base * 2

                # Poll sockets
                try:
                    while not owner.stop:
                        if not no_updates():
                            poller_register(broker_pub, POLLOUT)
                        activity = poller_poll(hb_period_base + int_(hb_period_spread * random_()))
                        if not activity:
                            broker_pub.send(HEARTBEAT_)
                            owner._idle()
                            break
                        tic_count -= 1
                        if tic_count < 0:
                            owner._tic()
                            tic_count = 100
                        # First send all sendable, empty the queue before trying to fill it
                        activity = dict_(activity)
                        if broker_pub in activity:
                            what = activity.pop(broker_pub)
                            if what & POLLOUT:
                                if not no_updates():
                                    send_update(broker_pub, noreply = T)
                                else:
                                    poller_unregister(broker_pub)
                        # Then put incoming stuff on the queue
                        for socket, what in activity.items():
                            if socket is pull:
                                pack = pull_recv_multipart()
                                if len_(pack) > 1:
                                    # ^ else Wakeup call, ignore
                                    put_nowait(pack)
                                elif pack[0] == "tic":
                                    owner._tic()
                                    tic_count = 100
                                del pack
                            elif socket is broker_rep and what & POLLIN:
                                owner._handle_update_request(broker_rep)
                except queue.Full:
                    self.logger.error("While handling IPSub FSM pipe: Queue full, update lost")
                except BootstrapNow:
                    self.transition(ZMQIPSub.FSM.Bootstrap)
                except:
                    self.logger.error("Exception in IPSub broker, re-bootstrapping in a sec", exc_info = True)
                    time.sleep(1)
                    self.transition(ZMQIPSub.FSM.Bootstrap)

            def leave(self):
                owner = self._owner()
                if owner is not None:
                    owner._notify_all(EVENT_LEAVE_BROKER, None)
                    owner._unbind()

    def __init__(self, broker_addresses, subscriptions = (), ctx=None):
        super(ZMQIPSub, self).__init__()

        self.broker_addresses = broker_addresses
        self.updates = queue.Queue(INPROC_HWM)
        self.current_update = None

        self.listener_req = self.listener_sub = None
        self.broker_rep = self.broker_pub = None
        self.local = threading.local()
        self._ndebug = None
        self._needs_subscriptions = True

        self.subscriptions = set(subscriptions)
        self.subscriptions.add(FRAME_HEARTBEAT)
        self.current_subscriptions = set()

        self.heartbeat_avg_period = 500
        self.heartbeat_push_timeout = 4000
        self.fsm_thread_id = None

        self.__context = ctx
        self.__context_lock = threading.Lock()

        self.reset()

    @property
    def context(self):
        if self.__context is None:
            with self.__context_lock:
                if self.__context is None:
                    self.__context = zmq.Context()
        return self.__context

    def reset(self):
        self.fsm = ZMQIPSub.FSM.Bootstrap(self)

    def run(self):
        # Must start in bootstrap
        assert not self.is_running

        # Initialize context by touching
        self.context

        self.stop = False
        while not self.stop:
            try:
                self.fsm_thread_id = thread.get_ident()
                self.fsm.enter()
                try:
                    while not self.stop:
                        self.fsm.stay()
                finally:
                    try:
                        self.fsm.leave()
                    except:
                        self.logger.error("Error cleaing up IPSub runner state", exc_info = True)
            except:
                if not self.stop:
                    self.logger.error("Uncaught exception in IPSub runner, resetting and relaunching", exc_info = True)
                self.reset()
            finally:
                self.fsm_thread_id = None

    def terminate(self):
        self.stop = True

    def _bind(self):
        ctx = self.context

        pub = ctx.socket(zmq.PUB)
        rep = ctx.socket(zmq.REP)
        set_hwm(pub, BROKER_PUB_HWM)
        set_hwm(rep, BROKER_REP_HWM)
        for addr in self.broker_addresses:
            pub_addr = addr.get('pub')
            rep_addr = addr.get('rep')
            if pub_addr:
                pub.bind(pub_addr)
            if rep_addr:
                rep.bind(rep_addr)

        self.broker_pub = pub
        self.broker_rep = rep

    def _repsocket(self):
        return self.broker_rep

    def _pubsocket(self):
        return self.broker_pub

    def _pullsocket(self):
        local = self.local
        if not hasattr(local, 'pull_socket'):
            ctx = self.context
            pull_socket = ctx.socket(zmq.PULL)
            set_hwm(pull_socket, INPROC_HWM)
            pull_socket.bind("inproc://IPSub%08x_queue" % id(self))
            local.pull_socket = pull_socket
        else:
            pull_socket = local.pull_socket
        return pull_socket

    def _pushsocket(self):
        local = self.local
        if not hasattr(local, 'push_socket'):
            ctx = self.context
            push_socket = ctx.socket(zmq.PUSH)
            set_hwm(push_socket, INPROC_HWM)
            push_socket.connect("inproc://IPSub%08x_queue" % id(self))
            local.push_socket = push_socket
        else:
            push_socket = local.push_socket
        return push_socket

    def _unbind(self):
        try:
            self.broker_pub.close()
        except:
            pass
        try:
            self.broker_rep.close()
        except:
            pass
        self.broker_pub = None
        self.broker_rep = None

    def _connect(self):
        ctx = self.context

        sub = ctx.socket(zmq.SUB)
        req = ctx.socket(zmq.REQ)
        set_hwm(sub, LISTENER_SUB_HWM)
        set_hwm(req, LISTENER_REQ_HWM)
        for addr in self.broker_addresses:
            pub_addr = addr.get('pub')
            rep_addr = addr.get('rep')
            if pub_addr:
                sub.connect(pub_addr)
            if rep_addr:
                req.connect(rep_addr)


        self.listener_sub = sub
        self.listener_req = req
        self.current_update = None
        return req

    def _disconnect(self):
        try:
            self.listener_sub.close()
        except:
            pass
        try:
            self.listener_req.close()
        except:
            pass
        self.listener_sub = None
        self.listener_req = None
        self.current_update = None

    def _subscribe(self):
        assert self.listener_sub is not None

        self._needs_subscriptions = False
        sub = self.listener_sub
        for prefix in self.subscriptions - self.current_subscriptions:
            sub.setsockopt(zmq.SUBSCRIBE, prefix)
        for prefix in self.current_subscriptions - self.subscriptions:
            sub.setsockopt(zmq.UNSUBSCRIBE, prefix)

        return sub

    def add_subscriptions(self, prefixes):
        self._needs_subscriptions = True
        self.subscriptions.update(prefixes)
        self.wake()

    def cancel_subscriptions(self, prefixes):
        self._needs_subscriptions = True
        self.subscriptions -= set(prefixes)
        self.wake()

    @property
    def has_updates(self):
        return not self.updates.empty()

    def _send_update(self, socket, noreply = False):
        if self.current_update is not None:
            # Waiting for a reply
            self.logger.error("IPSub FSM error: cannot send update when waiting for a reply, rebootstrapping")
            try:
                self.updates.put_nowait(self.current_update)
            except queue.Full:
                self.logger.error("While handling IPSub FSM error: Queue full, update lost")
            raise BootstrapNow

        try:
            update = self.updates.get_nowait()
        except queue.Empty:
            # So what...
            return

        socket.send_multipart(update)

        if not noreply:
            # Remember, we'll wait for a reply
            self.current_update = update

        # Notify listeners
        self._notify_all(EVENT_UPDATE_SENT, update)

    def _recv_update(self, socket):
        update = socket.recv_multipart()

        if not self._check_heartbeat(update):
            # Notify listeners
            self._notify_all(EVENT_INCOMING_UPDATE, update)
        else:
            self._idle()

    def _recv_update_reply(self, socket):
        # Check format without copying, assertion failures result in re-bootstrapping
        reply = socket.recv_multipart()

        if ( MIN_UPDATE_REPLY_FRAMES <= len(reply) <= MAX_UPDATE_REPLY_FRAMES
                 and len(reply[0]) <= MAX_UPDATE_REPLY_FIRSTFRAME ):
             reply_code = reply[0]
             if reply_code not in FRAME_VALID_UPDATE_REPLIES:
                 # Must be a reply payload, which implicitly means OK
                 reply_code = FRAME_UPDATE_OK
        else:
             reply_code = FRAME_UPDATE_OK

        # Notify listeners
        self._notify_all(EVENT_FOR_REPLY[reply_code], (self.current_update, reply))
        self.current_update = None

    def _check_heartbeat(self, update):
        HEARTBEAT_ = FRAME_HEARTBEAT

        # check for a heartbeat frame
        return (
            len(update) == 1
            and update[0] == HEARTBEAT_
        )

    def _handle_update_request(self, socket,
            isinstance=isinstance, BrokerReply=BrokerReply):
        update = socket.recv_multipart()
        if self._check_heartbeat(update):
            # Got a heartbeat, reply in kind
            socket.send(FRAME_HEARTBEAT)
            self._idle()
        else:
            # Real update, handle it
            try:
                self.updates.put_nowait(update)
                dropped = False
            except queue.Full:
                dropped = True

            # Notify listeners
            rv = self._notify_all(EVENT_INCOMING_UPDATE, update)
            if isinstance(rv, BrokerReply):
                socket.send_multipart(rv.reply)
            elif dropped:
                socket.send(FRAME_UPDATE_DROPPED)
            else:
                socket.send(FRAME_UPDATE_OK)

    def publish(self, prefix, payload, timeout = None, _ident = thread.get_ident):
        parts = [ prefix, self.identity ] + payload
        if _ident() == self.fsm_thread_id:
            try:
                self.updates.put_nowait(parts)
            except queue.Full:
                self.logger.error("While handling re-entrant IPSub publication: Queue full, update lost")
        else:
            push = self._pushsocket()
            if timeout is None:
                timeout = self.heartbeat_push_timeout
            if push.poll(timeout, zmq.POLLOUT):
                push.send_multipart(parts)
            else:
                self.logger.error("While handling IPSub publication: Push socket timeout, update lost")

    def wake(self):
        if self.__context is not None:
            try:
                # Don't block, mute means awaken and not catching up anyway
                push = self._pushsocket()
                if push.poll(1, zmq.POLLOUT):
                    push.send("")
            except zmq.ZMQError:
                # Shit happens, probably not connected
                pass

    @property
    def is_broker(self):
        return self.fsm.__class__ is ZMQIPSub.FSM.DesignatedBroker

    @property
    def is_running(self):
        return self.fsm.__class__ is not ZMQIPSub.FSM.Bootstrap




import functools
import weakref
import threading
import logging

import cython

from chorde.clients import base

cdef object CacheMissError, CancelledError, TimeoutError
cdef object wref, functools_partial
CacheMissError = base.CacheMissError
CancelledError = base.CancelledError
TimeoutError = base.TimeoutError
wref = weakref.ref
functools_partial = functools.partial

cdef bint strip_tracebacks
strip_tracebacks = False


def set_strip_tracebacks(bint strip):
    """ Sets traceback stripping behavior for async futures

    Async tracebacks have a tendency to cause memory leaks. Stripping them in
    production deployments is a good idea to avoid them, but keep them around
    during testing or non-production environments to get better diagnostic
    information and be able to detect those leaks.
    """
    global strip_tracebacks
    strip_tracebacks = strip


@cython.freelist(100)
cdef class ExceptionWrapper:
    cdef public object value
    cdef object __weakref__

    def __cinit__(self, value):
        self.value = value

    @cython.ccall
    def reraise(self, bint strip=True):
        exc = self.value
        if strip:
            del self.value
        try:
            exc_typ, exc_obj, exc_tb = exc
        finally:
            # Don't leave references to the exc/tb in the frame
            del exc
        try:
            if not strip:
                raise exc_typ(*exc_obj.args) from exc_obj
            elif exc_tb is not None:
                if exc_obj is not None:
                    if getattr(exc_obj, '__traceback__') is not exc_tb:
                        exc_obj = exc_obj.with_traceback(exc_tb)
                    raise exc_obj
                else:
                    raise exc_typ().with_traceback(exc_tb)
            elif exc_obj is not None:
                raise exc_obj
            else:
                raise exc_typ()
        finally:
            # Don't leave references to the exc/tb in the frame
            del exc_typ, exc_obj, exc_tb

@cython.freelist(100)
cdef class WeakCallback:
    cdef object me, callback, __weakref__

    def __cinit__(self, me, callback):
        self.me = wref(me)
        self.callback = callback

    def __call__(self, value):
        cdef object me
        me = self.me()
        if me is not None:
            return self.callback(me)

@cython.freelist(100)
cdef class DeferExceptionCallback:
    cdef object defer, __weakref__
    def __cinit__(self, defer):
        self.defer = defer
    def __call__(self, value):
        self.defer.set_exception(value[1] or value[0])

@cython.freelist(100)
cdef class ValueCallback:
    cdef object callback, __weakref__
    def __cinit__(self, callback):
        self.callback = callback
    def __call__(self, value):
        if value is not CacheMissError and not isinstance(value, ExceptionWrapper):
            return self.callback(value)

@cython.freelist(100)
cdef class MissCallback:
    cdef object callback, __weakref__
    def __cinit__(self, callback):
        self.callback = callback
    def __call__(self, value):
        if value is CacheMissError:
            return self.callback(value)

@cython.freelist(100)
cdef class ExceptionCallback:
    cdef object callback, __weakref__
    def __cinit__(self, callback):
        self.callback = callback
    def __call__(self, value):
        if isinstance(value, ExceptionWrapper):
            return self.callback(value)

@cython.freelist(100)
cdef class AnyCallback:
    cdef object on_value, on_miss, on_exc, __weakref__
    def __cinit__(self, on_value, on_miss, on_exc):
        self.on_value = on_value
        self.on_miss = on_miss
        self.on_exc = on_exc
    def __call__(self, value):
        if value is CacheMissError:
            if self.on_miss is not None:
                return self.on_miss()
        elif isinstance(value, ExceptionWrapper):
            if self.on_exc is not None:
                return self.on_exc((<ExceptionWrapper>value).value)
        else:
            if self.on_value is not None:
                return self.on_value(value)

@cython.freelist(100)
cdef class DoneCallback:
    cdef object callback, __weakref__
    def __cinit__(self, callback):
        self.callback = callback
    def __call__(self, value):
        return self.callback()

cdef class NONE:
    pass

@cython.freelist(100)
cdef class Future:
    def __cinit__(self, logger = None):
        self._cb = None
        self._value = NONE
        self._logger = logger
        self._done_event = None
        self._running = 0
        self._cancel_pending = 0
        self._cancelled = 0

    cpdef _set_nothreads(self, value):
        """
        Like set(), but assuming no threading is involved. It won't wake waiting threads,
        nor will it try to be thread-safe. Safe to call when the calling
        thread is the only one owning references to this future, and much faster.
        """
        cdef object old, cbs, _cb

        if self._value is not NONE:
            # No setting twice
            return

        # start atomic op
        old = self._value # avoid deadlocks due to finalizers
        _cb = self._cb
        if _cb is not None:
            self._cb = None
            self._value = value
            cbs = list(_cb) # end atomic op
        else:
            cbs = None
            self._value = value # end atomic op
        self._running = 0
        del old, _cb

        if cbs is not None:
            for cb in cbs:
                try:
                    cb(value)
                except:
                    if self._logger is not None:
                        error = self._logger
                    else:
                        error = logging.error
                    error("Error in async callback", exc_info = True)

    cpdef set(self, value):
        """
        Set the future's result as either a value, an exception wrappedn in ExceptionWrapper, or
        a cache miss if given CacheMissError (the class itself)
        """
        self._set_nothreads(value)

        if self._done_event is not None:
            # wake up waiting threads
            self._done_event.set()

    def set_result(self, value):
        self.set(value)

    def miss(self):
        """
        Shorthand for setting a cache miss result
        """
        self.set(CacheMissError)

    def _miss_nothreads(self):
        """
        Shorthand for setting a cache miss result without thread safety.
        See _set_nothreads
        """
        self._set_nothreads(CacheMissError)

    def exc(self, exc_info):
        """
        Shorthand for setting an exception result from an exc_info tuple
        as returned by sys.exc_info()
        """
        self.set(ExceptionWrapper.__new__(ExceptionWrapper, exc_info))

    def _exc_nothreads(self, exc_info):
        """
        Shorthand for setting an exception result from an exc_info tuple
        as returned by sys.exc_info(), without thread safety.
        See _set_nothreads
        """
        self._set_nothreads(ExceptionWrapper.__new__(ExceptionWrapper, exc_info))

    def set_exception(self, exception):
        """
        Set the Future's exception object.
        """
        if strip_tracebacks and getattr(exception, '__traceback__', None) is not None:
            exception = exception.with_traceback(None)
        self.exc((type(exception),exception,None))

    def on_value(self, callback):
        """
        When and if the operation completes without exception, the callback
        will be invoked with its result.
        """
        return self._on_stuff(ValueCallback.__new__(ValueCallback, callback))

    def on_miss(self, callback):
        """
        If the operation results in a cache miss, the callback will be invoked
        without arugments.
        """
        return self._on_stuff(MissCallback.__new__(MissCallback, callback))

    def on_exc(self, callback):
        """
        If the operation results in an exception, the callback will be invoked
        with an exc_info tuple as returned by sys.exc_info.
        """
        return self._on_stuff(ExceptionCallback.__new__(ExceptionCallback, callback))

    def on_any(self, on_value = None, on_miss = None, on_exc = None):
        """
        Handy method to set callbacks for all kinds of results, and it's actually
        faster than calling on_X repeatedly. None callbacks will be ignored.
        """
        return self._on_stuff(AnyCallback.__new__(AnyCallback, on_value, on_miss, on_exc))

    def on_any_once(self, on_value = None, on_miss = None, on_exc = None):
        """
        Like on_any, but will only set the callback if no other callback has been set
        """
        if self._cb is None:
            return self._on_stuff(AnyCallback.__new__(AnyCallback, on_value, on_miss, on_exc))

    def on_done(self, callback):
        """
        When the operation is done, the callback will be invoked without arguments,
        regardless of the outcome. If the operation is cancelled, it won't be invoked.
        """
        return self._on_stuff(DoneCallback.__new__(DoneCallback, callback))

    def chain(self, defer):
        """
        Invoke all the callbacks of the other defer
        """
        self._on_stuff(defer.set)

    def chain_std(self, defer):
        """
        Invoke all the callbacks of the other defer, without assuming the other
        defer follows our non-standard interface.
        """
        return self.on_any(
            defer.set_result,
            functools_partial(defer.set_exception, CacheMissError()),
            DeferExceptionCallback.__new__(DeferExceptionCallback, defer)
        )

    cpdef _on_stuff(self, callback):
        if self._value is NONE:
            if self._cb is None:
                self._cb = list()
            self._cb.append(callback)
        else:
            callback(self._value)
        return self

    def add_done_callback(self, callback):
        """
        When the operatio is done, the callback will be invoked with the
        future object as argument.
        """
        return self._on_stuff(WeakCallback.__new__(WeakCallback, self, callback))

    cdef int c_done(self) except -1:
        """
        Return True if the operation has finished, in a result or exception or cancelled, and False if not.
        """
        if self._value is not NONE or self._cancelled:
            return 1
        else:
            return 0

    def done(self):
        """
        Return True if the operation has finished, in a result or exception or cancelled, and False if not.
        """
        if self.c_done():
            return True
        else:
            return False

    def running(self):
        """
        Return True if the operation is running and cannot be cancelled. False if not running
        (yet or done).
        """
        if self._running:
            return True
        else:
            return False

    def cancelled(self):
        """
        Return True if the operation has been cancelled successfully.
        """
        if self._cancelled:
            return True
        else:
            return False

    def cancel_pending(self):
        """
        Return True if cancel was called.
        """
        if self._cancel_pending:
            return True
        else:
            return False

    def cancel(self):
        """
        Request cancelling of the operation. If the operation cannot be cancelled,
        it will return False. Otherwise, it will return True.
        """
        if self._cancelled:
            return False
        else:
            self._cancel_pending = 1
            return True

    def set_running_or_notify_cancelled(self):
        """
        To be invoked by executors before executing the operation. If it returns True,
        the operation may go ahead, and if False, a cancel has been requested and the
        operation should not be initiated, all threads waiting for the operation will
        be wakened immediately and the future will be marked as cancelled.
        """
        if self._cancel_pending:
            self._cancelled = 1
            self._running = 0

            # Notify waiters and callbacks
            self.set_exception(CancelledError())

            return False
        else:
            self._running = 1
            return True

    cdef c_result(self, timeout, int norecurse):
        """
        Return the operation's result, if any. If an exception was the result, re-raise it.
        If it was cancelled, raises CancelledError, and if timeout is specified and not None,
        and the specified time elapses without a result available, raises TimeoutError.
        """
        cdef object value
        cdef ExceptionWrapper exc_value

        if self._value is not NONE:
            value = self._value
            if isinstance(value, ExceptionWrapper):
                exc_value = <ExceptionWrapper>value
                try:
                    exc_value.reraise(False)
                finally:
                    del exc_value, value
            elif value is CacheMissError:
                raise CacheMissError()
            else:
                return value
        elif self._cancelled:
            raise CancelledError()
        else:
            if timeout is not None and timeout == 0:
                raise TimeoutError()
            else:
                # Wait for it
                if self._done_event is None:
                    self._done_event = threading.Event()
                if not norecurse and (self._value is not NONE or self._done_event.wait(timeout)):
                    return self.c_result(0, 1)
                elif self._cancelled:
                    raise CancelledError()
                else:
                    raise TimeoutError()

    def result(self, timeout=None, norecurse=False):
        """
        Return the operation's result, if any. If an exception was the result, re-raise it.
        If it was cancelled, raises CancelledError, and if timeout is specified and not None,
        and the specified time elapses without a result available, raises TimeoutError.
        """
        return self.c_result(timeout, norecurse)

    def exception(self, timeout=None):
        """
        If the operation resulted in an exception, return the exception object.
        Otherwise, return None. If the operation has been cancelled, raises CancelledError,
        and if timeout is specified and not None, and the specified time elapses without
        a result available, raises TimeoutError.
        """
        cdef object value
        cdef ExceptionWrapper exc_value

        if self._value is not NONE:
            value = self._value
            if isinstance(value, ExceptionWrapper):
                exc_value = <ExceptionWrapper>value
                return exc_value.value[1] or exc_value.value[0]
            elif value is CacheMissError:
                return CacheMissError
            else:
                return None
        elif self._cancelled:
            raise CancelledError()
        else:
            try:
                self.c_result(timeout, 0)
                return None
            except CancelledError:
                raise
            except Exception as e:
                try:
                    return e
                finally:
                    del e




# -*- coding: utf-8 -*-
import zmq

# Test zmq Frame's buffer protocol
try:
    buffer(zmq.Message(''))
    fbuffer = buffer
    bbytes = bytes
except TypeError:
    def fbuffer(frame):
        return frame.buffer if hasattr(frame, 'buffer') else buffer(frame)
    def bbytes(buf):
        return buf.tobytes() if hasattr(buf, 'tobytes') else bytes(buf)



# -*- coding: utf-8 -*-

""" Used to mock the integration between chorde and different frameworks/libraries
"""

def monkey_patch_tornado():
    """ Mock the tornado.conncurrent.is_future to mark the chorde
    Futures as valid.

    This is required, because Tornado checks that the result is a
    valid future, so if this isn't monkey patched when using
    chorde Futures is going to raise an exception
    """
    from tornado import concurrent
    from .clients.asyncache import Future
    original_futures = concurrent.FUTURES
    if isinstance(original_futures, tuple):
        if Future in original_futures:
            # already mocked so nothing to do
            return
        # on tornado 4.X it could be a tuple or just one class
        new_futures = original_futures + (Future,)
    else:
        new_futures = (original_futures, Future)

    concurrent.FUTURES = new_futures





# -*- coding: utf-8 -*-
from . import base, inproc
from .base import NONE, CacheMissError
from chorde import shmemutils, sPickle, serialize

import functools
import hashlib
import heapq
import json
import logging
import mmap
import os.path
import shutil
import tempfile
import _thread as thread
import threading
import time
import weakref

KEY_SUFFIX = '.key'
RAW_SUFFIX = '.raw'
UTF_SUFFIX = '.utf'
SER_SUFFIX = '.ser'
FILE_SUFFIX = '.file'
UNMAP_VALUE_SUFFIXES = (RAW_SUFFIX, UTF_SUFFIX, SER_SUFFIX)
VALUE_SUFFIXES = (FILE_SUFFIX,) + UNMAP_VALUE_SUFFIXES
ALL_SUFFIXES = VALUE_SUFFIXES + (KEY_SUFFIX,)

_caches_mutex = threading.RLock()
_caches = weakref.WeakKeyDictionary()

MMAP_THRESHOLD = 1 << 14

def _register_files(cache):
    _caches[cache] = None

def cachePurge(timeout = 0):
    with _caches_mutex:
        caches = list(_caches.keys())

    for cache in caches:
        cache.purge(timeout)

def cacheClear():
    with _caches_mutex:
        caches = list(_caches.keys())

    for cache in caches:
        cache.clear()

class CacheJanitorThread(threading.Thread):

    def __init__(self, sleep_interval, purge_timeout = 0):
        threading.Thread.__init__(self)
        self.sleep_interval = sleep_interval
        self.purge_timeout = purge_timeout
        self.logger = None
        self.setDaemon(True)

    def run(self):
        global cachePurge

        while True:
            time.sleep(self.sleep_interval)
            try:
                cachePurge(self.purge_timeout)
            except:
                if self.logger is None:
                    pass
                else:
                    self.logger.error("Exception during cache purge", exc_info = True)

def startCacheJanitorThread(sleep_interval=3600, purge_timeout=0):
    jthread = CacheJanitorThread(sleep_interval, purge_timeout)
    jthread.start()
    return jthread

def _tmpsuffix():
    return ".tmp.%d.%x.%x" % (os.getpid(),thread.get_ident(),int(time.time()))

def _touch(path):
    try:
        os.utime(path, (time.time(), os.path.getmtime(path)))
    except OSError:
        pass

def _nodelete(f):
    f.delete = False
    if hasattr(f, '_closer'):
        f._closer.delete = False

def _swap(source, dest, sizeback = None):
    cursize = 0
    if sizeback is not None:
        # remember the existing file's size
        try:
            cursize = os.path.getsize(dest)
        except:
            pass
    try:
        os.rename(source, dest)
    except OSError:
        # Do an indirect swap, in case it was on a different filesystem
        tmpname = dest+_tmpsuffix()
        try:
            try:
                # This might work still, on systems with os.link where rename doesn't overwrite
                os.link(source, tmpname)
            except:
                shutil.copy2(source, tmpname)
            try:
                # Try again the rename, if the above did the copy, it might work now
                os.rename(tmpname, dest)
            except:
                # Non-atomic replace, needed in windows
                os.unlink(dest)
                os.rename(tmpname, dest)
            os.unlink(source)
        except:
            _clean(tmpname)
            raise

    if sizeback is not None:
        # Call with size delta
        sizeback(os.path.getsize(dest) - cursize)

def _link(source, dest, sizeback = None, filemode = None):
    cursize = 0
    if sizeback is not None:
        # remember the existing file's size
        try:
            cursize = os.path.getsize(dest)
        except:
            pass
    try:
        os.link(source, dest)
    except OSError:
        # Do an indirect swap, in case it was on a different filesystem
        tmpname = dest+_tmpsuffix()
        try:
            shutil.copy2(source, tmpname)
            if filemode is not None:
                os.chmod(tmpname, filemode)
            try:
                # Try again the rename, it might work now
                os.rename(tmpname, dest)
            except:
                # Non-atomic replace, needed in windows
                os.unlink(dest)
                os.rename(tmpname, dest)
        except:
            _clean(tmpname)
            raise

    if sizeback is not None:
        # Call with size delta
        sizeback(os.path.getsize(dest) - cursize)

def _clean(path, sizeback = None):
    if sizeback is not None:
        # remember size, for the callback
        try:
            sizeback = functools.partial(sizeback, os.path.getsize(path))
        except:
            # Doesn't exist...
            return

    try:
        os.unlink(path)
    except:
        pass

    if sizeback is not None:
        sizeback()

def _clean_other_values(path, keep_suffix, sizeback = None):
    for suffix in VALUE_SUFFIXES:
        if suffix != keep_suffix:
            _clean(path+suffix, sizeback)

def json_dumpb(obj):
    return json.dumps(obj).encode('ascii')

class FilesCacheClient(base.BaseCacheClient):
    """
    This cache client will store itmes on a file hierarchy.
    Depending on the operating system, the files need not be backed by
    persistent storage (ie: it could be temporary in-memory storage, making
    it quite fast), but persistence is a possibility, and a common reason to use
    this client.

    This client will behave specially when given file, bytes, bytearray and memoryview
    objects. Normal clients can't or shouldn't store file objects, but this client
    has been specifically designed for that task.

    File objects thus cached must contain a 'name' attribute that
    points to an absolute path to the file.

    When a file object is added to the client, it is copied in the most efficient
    manner into the cache's file heirarchy. On posix systems, this will create
    a hardlink when the file is on the same filesystem as the cache's store,
    and it is thus the recommended use case.

    When byte strings are given, they will be written to a file within the cache's
    file heirarchy. Retrieving them is transparent, unless mmap_raw is given and True.
    In that case, queries to the cache will return a buffer (maybe a string, maybe
    an mmap object, depending on the value's size), and thus the cache will not be
    transparent, but it will be more efficient as it will share memory across processes.

    Bytearray and memoryview objects are treated similarly.

    This cache is limited by size in bytes. The size includes book-keeping overhead,
    so be certain to account for serialized key sizes when specifying it.

    The client is thread and process-safe, but it needs a purging thread to be
    running, either manually calling purge() every once in a while, or by calling
    startCacheJanitorThread() with appropriate arguments.

    If entries are added too fast for the purging thread, that is, if considerably
    more bytes than the limit are added between purging runs, it is possible for this
    cache to grow past the size limit. If this is a concern, specify sync_purge as
    the highest acceptable multiple of the cache's size limit. When reached, a
    synchronous purge will be attempted before returning, guaranteeing adherence
    to given size limits.
    """

    def __init__(self, size, basepath,
            failfast_size = 500, failfast_time = 0.25, counter_slots = 256,
            key_pickler = json_dumpb, value_pickler = None, value_unpickler = None, value_opener=None, checksum_key = None,
            dirmode = 0o700, filemode = 0o400, mmap_raw = False,
            sync_purge = None):
        """
        Params:
            size: Maximum cache size, in bytes, counting key serialization and book-keeping overheads

            failfast_size: Entries in a small key-only inproc cache used to speed up cache misses

            failfast_time: TTL of the failfast cache

            counter_slots: Number of PID slots in the cached size counter file. In order to make size
                updates wait-free, each process uses a separate entry. In platforms where fcntl.lock
                is implemented, this needs only be as big as the number of processes that will attempt
                to concurrently update this cache, plus some free space to allow dying processes to
                respawn (their slot will be unusable). In other platforms, this should be 65536, since
                otherwise hash collisions would prevent the cache from starting (there's no atomic
                way of grabbing a counter slot implemented in those platforms). The default should
                work on most posix platforms.

            key_pickler: Pickling *function* used to serialize keys. By default, it's json. This
                function should be stable: equal keys must serialize to equal strings.
                Important Note: Pickle doesn't always respect this invariant.

            value_pickler: Pickling *function* used to serialize values. It should take a signature
                compatible with pickle.dump, including protocol version. By default, it's built
                out of sPickle.dump and the checksum_key (which is then mandatory)

            value_unpickler: Unpickling *function* used to de-serialize values. It should take a
                signature compatible with pickle.load. By default, it's built out of sPickle.load
                and the checksum_key (which is then mandatory). Always validate contents if this
                function is based on pickling, otherwise malicious injection into the cache
                file heirarchy could result in arbitrary code execution.

            value_opener: opener *function* used to return non-file object. The function must expect two
                arguments:
                    * filename (str): The complete path where the file is located.
                    * mode (str): The same modes expected by built-in `open` function.

            checksum_key: When using the default picklers, this *private* key is necessary
                in order to authenticate values and make sure they have been written by this process
                (or any process with access to the key). Otherwise, malicious injection into the
                cache's file heirarchy could result in arbitrary code execution, and file corruption
                even if not malicious, could result in hard crashes (out-of-memory, segmentation fault, etc).

            dirmode: Permissions bitfield used for creating directories. See os.chmod

            filemode: Permissions bitfield used for cache files. See os.chmod

            mmap_raw: Whether to return mmap objects when reading large byte strings from the cache,
                instead of string copies of the contents. This is far more efficient than reading the
                whole contents of the file, but the returned buffer isn't fully compatible with string semantics,
                so by default it's False.

            sync_purge: A (possibly float) multiplier of size that will trigger a synchronous purge when
                reached. To avoid the long pauses that could be generated by those forced purge cycles, this
                is None (no sync purges) by default.
        """
        if isinstance(checksum_key, str):
            checksum_key = checksum_key.encode("utf8")
        self._failfast_cache = inproc.Cache(failfast_size)
        self.failfast_time = failfast_time
        self.basepath = basepath
        self.max_size = size
        self.counter_slots = counter_slots
        self.size = self._make_counter()
        self.key_pickler = key_pickler
        self.checksum_key = checksum_key
        self.filemode = filemode
        self.dirmode = dirmode
        self.mmap_raw = mmap_raw
        self.sync_purge = sync_purge

        if value_opener is None:
            value_opener = open
        self.value_opener = value_opener

        if value_pickler is None or value_unpickler is None:
            if checksum_key is None:
                raise ValueError("Must be given a checksum_key when using the default pickler")
            else:
                self.value_pickler = functools.partial(sPickle.dump, checksum_key)
                self.value_unpickler = functools.partial(sPickle.load, checksum_key)
        elif value_pickler is not None or value_unpickler is not None:
            raise ValueError("Must be given both pickler and unpickler when using custom picklers")
        else:
            self.value_pickler = value_pickler
            self.value_unpickler = value_unpickler

        _register_files(self)

    def _make_counter(self):
        counter_slots = self.counter_slots
        if self.max_size > (1<<30):
            cls = shmemutils.SharedCounter64
            basename = "sizemap.64"
        else:
            cls = shmemutils.SharedCounter32
            basename = "sizemap.32"
        path = os.path.join(self.basepath, "%s.%x" % (basename, counter_slots))

        try:
            rv = cls.from_path(counter_slots, path)
        except:
            # ouch, broken counter, reset

            # Create a new one
            tpath = path + _tmpsuffix()
            rv = cls.from_path(counter_slots, tpath)
            rv.close()

            # Swap it
            _swap(tpath, path)

            # Re-open it
            rv = cls.from_path(counter_slots, path)
        if int(rv) == 0:
            logger = logging.getLogger('chorde.files')

            t0 = time.time()
            logger.info("Initializing concurrent counter for file-based cache on %r", self.basepath)

            # Um... uninitialized
            # Compute size
            actual_size = self._compute_size()

            # Computing takes time, check to see whether some other guy beat us to it
            if int(rv) == 0:
                # Nop, set
                rv += actual_size
                rv.flush()

            t1 = time.time()
            logger.info("Initialized concurrent counter for file-based cache on %r (%.3fs)",
                self.basepath, t1 - t0)

        return rv

    def _walk(self):
        for dirpath, dirs, files in os.walk(self.basepath):
            if dirpath != self.basepath:
                for fname in files:
                    yield os.path.join(dirpath, fname)

    def _compute_size(self):
        rv = 0
        getsize = os.path.getsize
        join = os.path.join
        for fpath in self._walk():
            try:
                rv += getsize(join(self.basepath, fpath))
            except OSError:
                pass # Meh, entries are transient
        return rv

    def _mkpath(self, key):
        khash = hashlib.md5(key).hexdigest()
        return (khash[:2], khash[2:4], khash[4:6], khash[6:])

    def _mktmp(self, **kw):
        tmpdir = os.path.join(self.basepath, "staging")
        if not os.path.exists(tmpdir):
            try:
                os.makedirs(tmpdir, self.dirmode)
            except:
                if not os.path.exists(tmpdir):
                    raise
        return tempfile.NamedTemporaryFile(dir=tmpdir, **kw)

    @property
    def is_async(self):
        return False

    @property
    def capacity(self):
        return self.max_size

    @property
    def usage(self):
        return int(self.size)

    def _put(self, key, value, ttl, replace):
        key = self.key_pickler(key)
        kpath = self._mkpath(key)

        targetpath = os.path.join(self.basepath, *kpath)
        keypath = targetpath + KEY_SUFFIX
        targetdir = os.path.join(self.basepath, *kpath[:-1])

        if not os.path.exists(targetdir):
            try:
                os.makedirs(targetdir, self.dirmode)
            except OSError:
                if not os.path.exists(targetdir):
                    raise

        reuse_keyfile = False
        if os.path.exists(keypath):
            # Huh... compare
            if os.path.getsize(keypath) == len(key):
                with open(keypath, "rb") as ekey:
                    if ekey.read() == key:
                        reuse_keyfile = True
            if reuse_keyfile and not replace:
                # Um... check the validity of the current value before going further
                if os.path.getmtime(keypath) >= time.time():
                    # Still valid, don't replace
                    return False

        def setTtl(path):
            now = time.time()
            os.utime(path, (now, now + ttl))

        def setPermsf(fileobj):
            os.fchmod(fileobj.fileno(), self.filemode)

        if not reuse_keyfile:
            # Create a key file
            # Yeah, can't use a context manager, pity
            keyfile = self._mktmp()
            try:
                keyfile.write(key)
                keyfile.flush()
            except:
                keyfile.close()
                raise

            # On posix systems, where swap does an atomic rename,
            # which preserves source attributes, this avoids instants
            # where the entry is there but expired. Its TTL will be reset
            # with the proper time from insertion in the cache later on.
            setTtl(keyfile.name)
            setPermsf(keyfile)

        try:
            try:
                if hasattr(value, 'fileno') and hasattr(value, 'name') and os.path.isabs(value.name):
                    # Looks like a file, so we can simply link, try
                    name = value.name

                    # Overwiting from unknown location, must take extra care
                    _link(name, targetpath+FILE_SUFFIX, self.size.__iadd__, self.filemode)
                    if not reuse_keyfile:
                        _swap(keyfile.name, keypath, self.size.__iadd__)
                        _nodelete(keyfile)

                    # Succeeded, clean up other representations, if they exist, set ttl
                    _clean_other_values(targetpath, FILE_SUFFIX, self.size.__isub__)
                    setTtl(keypath)

                    return True
            except:
                logging.error("Oops", exc_info = True)
                # Meh, linking didn't work
                pass

            # Else, must stuff the values into a file
            if isinstance(value, (bytes, memoryview, bytearray)):
                # Quite easily... stuff it
                with self._mktmp() as rawfile:
                    rawfile.write(value)
                    rawfile.flush()

                    _swap(rawfile.name, targetpath+RAW_SUFFIX, self.size.__iadd__)
                    _nodelete(rawfile)
                    if not reuse_keyfile:
                        _swap(keyfile.name, keypath, self.size.__iadd__)
                        _nodelete(keyfile)

                    # Succeeded, clean up other representations, if they exist
                    _clean_other_values(targetpath, RAW_SUFFIX, self.size.__isub__)
                    setTtl(keypath)

                return True
            elif isinstance(value, (str,)):
                # Quite easily... stuff it
                with self._mktmp(mode='w+', encoding="utf8") as rawfile:
                    rawfile.write(value)
                    rawfile.flush()

                    _swap(rawfile.name, targetpath+UTF_SUFFIX, self.size.__iadd__)
                    _nodelete(rawfile)
                    if not reuse_keyfile:
                        _swap(keyfile.name, keypath, self.size.__iadd__)
                        _nodelete(keyfile)

                    # Succeeded, clean up other representations, if they exist
                    _clean_other_values(targetpath, UTF_SUFFIX, self.size.__isub__)
                    setTtl(keypath)

                return True
            else:
                if not self.checksum_key:
                    raise RuntimeError("Cannot encode arbitrary objects without a checksum key")

                with self._mktmp() as rawfile:
                    self.value_pickler(value, rawfile, 2)
                    rawfile.flush()

                    _swap(rawfile.name, targetpath+SER_SUFFIX, self.size.__iadd__)
                    _nodelete(rawfile)
                    if not reuse_keyfile:
                        _swap(keyfile.name, keypath, self.size.__iadd__)
                        _nodelete(keyfile)

                    # Succeeded, clean up other representations, if they exist
                    _clean_other_values(targetpath, SER_SUFFIX, self.size.__isub__)
                    setTtl(keypath)

                return True
        finally:
            if not reuse_keyfile:
                keyfile.close()

        return True

    def put(self, key, value, ttl):
        self._put(key, value, ttl, True)

        try:
            del self._failfast_cache[key]
        except:
            pass

        if self.sync_purge is not None and int(self.size) > (self.max_size * self.sync_purge):
            # Ehm... going overboard man
            try:
                self._do_purge(0)
            except serialize.DeadlockError:
                # Someone's purging already
                pass

    def renew(self, key, ttl):
        key = self.key_pickler(key)
        kpath = self._mkpath(key)

        targetpath = os.path.join(self.basepath, *kpath)
        keypath = targetpath + KEY_SUFFIX
        targetdir = os.path.join(self.basepath, *kpath[:-1])

        if not os.path.exists(targetdir):
            return

        if os.path.exists(keypath):
            # Huh... compare
            if os.path.getsize(keypath) == len(key):
                with open(keypath, "rb") as ekey:
                    if ekey.read() == key:
                        # Touch
                        now = time.time()
                        kttl = os.path.getmtime(keypath)
                        nttl = ttl + now
                        if kttl < nttl:
                            os.utime(keypath, (now, now + ttl))

    def add(self, key, value, ttl):
        rv = self._put(key, value, ttl, False)

        try:
            del self._failfast_cache[key]
        except:
            pass

        if rv and self.sync_purge is not None and int(self.size) > (self.max_size * self.sync_purge):
            # Ehm... going overboard man
            try:
                self._do_purge(0)
            except serialize.DeadlockError:
                # Someone's purging already
                pass

        return rv

    def delete(self, key):
        key = self.key_pickler(key)
        kpath = self._mkpath(key)
        targetpath = os.path.join(self.basepath, *kpath)

        for suffix in ALL_SUFFIXES:
            _clean(targetpath+suffix, self.size.__isub__)

    def expire(self, key):
        key = self.key_pickler(key)
        kpath = self._mkpath(key)
        targetpath = os.path.join(self.basepath, *kpath)

        # Touch all files, invalidating them immediately
        for suffix in ALL_SUFFIXES:
            try:
                now = time.time()
                os.utime(targetpath+suffix, (now,now))
            except:
                pass

    def _getTtl(self, key, default = base.NONE, baseNONE = base.NONE, ttl_skip=None,
            promote_callback = None, decode=True, return_stale=False, _ninf=float('-inf')):
        key = self.key_pickler(key)
        kpath = self._mkpath(key)
        targetpath = os.path.join(self.basepath, *kpath)
        keypath = targetpath + KEY_SUFFIX

        if not os.path.exists(keypath):
            return default, -1

        try:
            if os.path.getsize(keypath) != len(key):
                return default, -1
        except OSError:
            return default, -1

        # Check fail-fast cache before opening files and all taht
        if self._failfast_cache.get(key, _ninf) > (time.time() - self.failfast_time):
            return default, -1

        with open(keypath, "rb") as ekey:
            if ekey.read() == key:
                # Um... check the validity of the current value before going further
                ttl = os.path.getmtime(keypath)
                rttl = ttl - time.time()

                if not return_stale and (ttl_skip is not None and rttl < ttl_skip):
                    return default, -1
                elif decode:
                    try:
                        _touch(keypath) # force atime update, in case it's a non-strict-atime mount
                        if os.access(targetpath+FILE_SUFFIX, os.R_OK):
                            return self.value_opener(targetpath+FILE_SUFFIX, 'rb') ,rttl
                        elif os.access(targetpath+RAW_SUFFIX, os.R_OK):
                            with open(targetpath+RAW_SUFFIX, 'rb') as rawfile:
                                rawfile.seek(0, os.SEEK_END)
                                size = rawfile.tell()
                                rawfile.seek(0)
                                if size > MMAP_THRESHOLD:
                                    return mmap.mmap(rawfile.fileno(), size, access = mmap.ACCESS_READ), rttl
                                elif size == 0:
                                    return b"", rttl
                                else:
                                    return rawfile.read(), rttl
                        elif os.access(targetpath+UTF_SUFFIX, os.R_OK):
                            with open(targetpath+UTF_SUFFIX, 'r', encoding="utf8") as rawfile:
                                return rawfile.read(), rttl
                        elif self.checksum_key and os.access(targetpath+SER_SUFFIX, os.R_OK):
                            with open(targetpath+SER_SUFFIX, 'rb') as rawfile:
                                try:
                                    return self.value_unpickler(rawfile), rttl
                                except Exception:
                                    # Remove the broken entry from the cache
                                    logging.getLogger('chorde.files').warning(
                                        "Purging badly serialized entry from cache")
                                    for suffix in ALL_SUFFIXES:
                                        _clean(targetpath+suffix, self.size.__isub__)
                                    raise
                        else:
                            # Broken?
                            return default, -1
                    except:
                        # Oops
                        logging.getLogger('chorde.files').error("Error retrieving file contents", exc_info = True)
                        return default, -1
                else:
                    return default, rttl
            else:
                self._failfast_cache[key] = time.time()
                return default, -1

    def getTtl(self, key, default=NONE, ttl_skip = None, **kw):
        # This trampoline is necessary to avoid re-entrancy issues when this client
        # is wrapped inside a SyncWrapper. Internal calls go directly to _getTtl
        # to avoid locking the wrapper's mutex.
        return self._getTtl(key, default, ttl_skip = ttl_skip, return_stale = True, **kw)

    def get(self, key, default=NONE, **kw):
        rv, ttl = self._getTtl(key, default, ttl_skip = 0, **kw)
        if ttl < 0 and default is NONE:
            raise CacheMissError(key)
        else:
            return rv

    def contains(self, key, ttl = None):
        rv, ettl = self._getTtl(key, ttl_skip = 0, decode = False, return_stale = False)
        if ettl < 0:
            return False
        else:
            # Key check was successful, so... forget about past mistakes.
            try:
                del self._failfast_cache[key]
            except:
                pass
            return ttl is None or ettl > ttl

    def close(self):
        # Free up stuff
        self.size.close()
        self.size = None

    def clear(self):
        # Bye bye everything
        self.size.flush()
        for dirpath, dirnames, filenames in os.walk(self.basepath):
            for subpath in dirnames:
                try:
                    shutil.rmtree(os.path.join(dirpath, subpath))
                except OSError:
                    logging.error("Oops", exc_info = True)
                    pass
            for subpath in filenames:
                try:
                    os.unlink(os.path.join(dirpath, subpath))
                except OSError:
                    logging.error("Oops", exc_info = True)
                    pass
            del dirnames[:]
            del filenames[:]

        # Must reset counter now
        self.size.close()
        self.size = self._make_counter()
        self._failfast_cache.clear()

    def purge(self, timeout = 0):
        # Trampoline necessary to avoid deadlocks if this is wrapped in a SyncWrapper
        self._do_purge(timeout)

    @serialize.serialize
    def _do_purge(self, timeout):
        # Blocking trampoline
        self.__do_purge(timeout)

    @serialize.serialize(deadlock_timeout=0.01)
    def _try_purge(self, timeout):
        # Non-blocking trampoline
        self.__do_purge(timeout)

    def __do_purge(self, timeout):
        # Abbreviations to make it more readable
        exists = os.path.exists
        join = os.path.join
        getsize = os.path.getsize
        basepath = self.basepath

        # Since we're at it, compute size
        fullsize = 0
        expired_items = expired_bytes = evicted_items = evicted_bytes = 0

        logger = logging.getLogger('chorde.files')
        logger.info('Purging file-based cache at %r, cached size %d / %d',
            basepath, int(self.size), self.max_size)

        if int(self.size) > self.max_size:
            # Dingbats... must remove some entries
            deletions = []
            excess = int(self.size) - self.max_size
        else:
            # Don't even bother gathering deletions
            deletions = None
            excess = 0

        expirations = []

        for fpath in self._walk():
            try:
                fullsize += getsize(join(basepath, fpath))
            except OSError:
                pass
            if fpath.endswith(KEY_SUFFIX):
                bpath = fpath[:-4]
                fpath = join(basepath, fpath)
                bpath = join(basepath, bpath)
                try:
                    stats = os.stat(fpath)
                except OSError:
                    continue
                if (stats.st_mtime + timeout) < time.time():
                    # Stale key, say bye
                    expirations.append(bpath)
                elif deletions is not None:
                    itemsize = stats.st_size
                    for suffix in ALL_SUFFIXES:
                        try:
                            itemsize += os.path.getsize(bpath+suffix)
                        except OSError:
                            pass
                    delitem = (stats.st_atime, bpath, itemsize)
                    if deletions and excess <= 0:
                        if deletions[0] > delitem:
                            excess -= itemsize
                            xdelitem = heapq.heapreplace(deletions, delitem)
                            excess += xdelitem[2]
                    else:
                        excess -= itemsize
                        heapq.heappush(deletions, delitem)
            else:
                for suffix in VALUE_SUFFIXES:
                    if fpath.endswith(suffix):
                        bpath = fpath[:-len(suffix)]
                        break
                else:
                    bpath = None
                if bpath is not None:
                    fpath = join(basepath, fpath)
                    bpath = join(basepath, bpath)
                    if not exists(bpath+KEY_SUFFIX):
                        # Deleted entry
                        delta = []
                        _clean(fpath, delta.append)
                        if delta:
                            delta = sum(delta)
                            self.size -= delta
                            fullsize -= delta
                            excess -= delta
                            logger.debug('Removed stray file %r of size %d', fpath, delta)
                else:
                    # Extraneous file, remove
                    delta = []
                    _clean(join(basepath,fpath), delta.append)
                    if delta:
                        delta = sum(delta)
                        self.size -= delta
                        fullsize -= delta
                        excess -= delta
                        logger.debug('Removed extraneous file %r of size %d', fpath, delta)

        if expirations:
            for bpath in expirations:
                delta = []
                for suffix in ALL_SUFFIXES:
                    _clean(bpath+suffix, delta.append)
                if delta:
                    delta = sum(delta)
                    self.size -= delta
                    excess -= delta
                    fullsize -= delta
                    expired_items += 1
                    expired_bytes += delta
                    logger.debug('Removed stale entry of size %d at %r', delta, bpath)

        if deletions:
            while deletions and excess <= -deletions[0][2]:
                delitem = heapq.heappop(deletions)
                excess += delitem[2]
            for atime,bpath,itemsize in deletions:
                delta = []
                for suffix in ALL_SUFFIXES:
                    _clean(bpath+suffix, delta.append)
                if delta:
                    delta = sum(delta)
                    self.size -= delta
                    fullsize -= delta
                    evicted_items += 1
                    evicted_bytes += delta
                    logger.debug('Removed overflow entry of size %d LRU %fs ago at %r',
                        delta, time.time() - atime, bpath)

        self._failfast_cache.clear()

        logger.info('Expired %d bytes in %d items', expired_bytes, expired_items)
        logger.info('Evicted %d bytes in %d items', evicted_bytes, evicted_items)

        # We've been computing the actual size, so adjust the approximation
        adjustment = fullsize - int(self.size)
        if adjustment != 0:
            self.size += adjustment
            logger.info('Adjusted size discrepancy of %+d (final %d)', adjustment, int(self.size))




# -*- coding: utf-8 -*-
from threading import RLock, Lock, Condition
import time
import logging
import sys

from weakref import WeakValueDictionary
from functools import wraps

__all__ = [
    "Lock",
    "RLock",
    "Condition",
    "ReadWriteLock",
    "ReadWriteRLock",
]


class ReadWriteLock(object):
    __slots__ = [ '_read_ready', '_readers' ]

    locktype = staticmethod(Lock)

    def __init__(self):
        self._read_ready = Condition(self.locktype())
        self._readers = 0

    def acquire_read(self, blocking=1):
        if not self._read_ready.acquire(blocking):
            return False
        try:
            self._readers += 1
            return True
        finally:
            self._read_ready.release()

    def release_read(self):
        self._read_ready.acquire()
        try:
            self._readers -= 1
            if self._readers == 0:
                self._read_ready.notifyAll()
        finally:
            self._read_ready.release()

    def acquire_write(self, blocking=1):
        if not self._read_ready.acquire(blocking):
            return False
        while self._readers > 0:
            self._read_ready.wait(blocking or None)
            if not blocking and self._readers > 0:
                self._read_ready.release()
                return False
        return True

    def release_write(self):
        self._read_ready.release()

class ReadWriteRLock(ReadWriteLock):
    __slots__ = []
    locktype = staticmethod(RLock)

def serialize_docstring(f):
    if 'sync' not in (getattr(f, '__doc__', '') or '').lower():
        f.__doc__ = 'Synchronized\n\n' + (getattr(f, '__doc__', '') or '')
    return f

_initialization_lock = Lock()

def serialize(*p,**kw):
    if 'pydoc' in sys.argv[0]:
        if len(p) == 1:
            return serialize_docstring(*p)
        else:
            return serialize_docstring

    acquire   = kw.get('acquire','acquire')
    release   = kw.get('release','release')

    if 'lockobj' in kw:
        # pre-assigned lock object, unkeyed
        lockobj = kw['lockobj']
        def decor(f):
            @wraps(f)
            def rv(*pos,**kw):
                try:
                    getattr(lockobj, acquire)()
                    return f(*pos,**kw)
                finally:
                    getattr(lockobj, release)()
            return rv
    else:
        # unkeyed lock attribute
        locktype = kw.get('locktype',RLock)
        lockattr = kw.get('lockmember','__serialization_lock')
        def decor(f):  # lint:ok
            @wraps(f)
            def rv(self,*pos,**kw):
                if not hasattr(self,lockattr):
                    with _initialization_lock:
                        if not hasattr(self,lockattr):
                            setattr(self,lockattr, locktype())
                lock = getattr(self,lockattr)
                try:
                    getattr(lock, acquire)()
                    return f(self,*pos,**kw)
                finally:
                    getattr(lock, release)()
            return rv

    if len(p) == 1:
        return decor(p[0])
    else:
        return decor

def serialize_read(*p,**kw):
    kw['locktype']= kw.get('locktype', ReadWriteRLock)
    kw['acquire'] = kw.get('acquire', 'acquire_read')
    kw['release'] = kw.get('release', 'release_read')
    return serialize(*p,**kw)

def serialize_write(*p,**kw):
    kw['locktype']= kw.get('locktype', ReadWriteRLock)
    kw['acquire'] = kw.get('acquire', 'acquire_write')
    kw['release'] = kw.get('release', 'release_write')
    return serialize(*p,**kw)



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod, abstractproperty
import operator
from chorde.serialize import serialize_read, serialize_write, serialize

# Overridden by inproc_cache based on LRUCache availability
CacheMissError = KeyError

class NoServersError(Exception):
    pass

try:
    from concurrent.futures import TimeoutError, CancelledError
except ImportError:
    class TimeoutError(Exception):  # lint:ok
        pass
    class CancelledError(Exception):  # lint:ok
        pass

class NONE: pass
# 100 years ttl
FOREVER =  3600 * 24 * 30 * 12 * 100

class BaseCacheClient(object):
    """
    Interface of all backing stores.
    """
    __metaclass__ = ABCMeta

    @abstractproperty
    def is_async(self):
        return False

    @abstractproperty
    def capacity(self):
        raise NotImplementedError

    @abstractproperty
    def usage(self):
        raise NotImplementedError

    def wait(self, key, timeout = None):
        """
        Only valid for async clients, if there is a write pending, this
        will wait for it to finish. If there is not, it will return immediately.
        """
        pass

    @abstractmethod
    def put(self, key, value, ttl):
        raise NotImplementedError

    @abstractmethod
    def renew(self, key, ttl):
        """
        If there is a value for that key in the cache, and if its TTL is lower
        than the specified value, it will change its TTL in an atomic fashion
        and extend the life of that value.
        """
        raise NotImplementedError

    def add(self, key, value, ttl):
        """
        Like put, sets a value, but add only does so if there wasn't a valid
        value before, and does so as atomically as possible (ie: it tries to
        be atomic, but doesn't guarantee it - see specific client docs).

        Returns True when it effectively stores the item, False when it
        doesn't (there was one before)
        """
        if not self.contains(key):
            self.put(key, value, ttl)
            return True
        else:
            return False

    @abstractmethod
    def delete(self, key):
        raise NotImplementedError

    def expire(self, key):
        """
        If they given key is in the cache, it will set its TTL to 0, so
        getTtl will subsequently return either a miss, or an expired item.
        """
        return self.delete(key)

    @abstractmethod
    def getTtl(self, key, default = NONE, ttl_skip = None, promote_callback = None):
        """
        Returns: a tuple (value, ttl). If a default is given and the value
            is not in cache, return (default, -1). If a default is not given
            and the value is not in cache, raises CacheMissError. If the value
            is in the cache, but stale, ttl will be < 0, and value will be
            other than NONE. Note that ttl=0 is a valid and non-stale result.
            If ttl_skip is given, and the cache is an aggregate of multiple
            caches, entries with ttl below ttl_skip will be ignored, and
            automatic promotion of remaining entries will occur.

            If promote_callback is given, and the operation results in a promotion
            from a lower level to a higher one, promote_callback will be called
            with 3 arguments: the key as given, the value being promoted,
            and the remaining TTL.
        """
        if default is NONE:
            raise CacheMissError(key)
        else:
            return (default, -1)

    def promote(self, key, default = NONE, ttl_skip = None, promote_callback = None):
        """
        Returns: None. If getTtl would find an entry with the given arguments,
            this entry will be promoted. Otherwise, it's a no-op.
        """
        return

    def get(self, key, default = NONE, **kw):
        rv, ttl = self.getTtl(key, default, **kw)
        if ttl < 0:
            if default is NONE:
                raise CacheMissError(key)
            else:
                return default
        else:
            return rv

    def getMulti(self, keys, default = NONE, **kw):
        """
        Params:

            keys: An iterable of keys to fetch

            default: The default value to return when a miss occurs

            ttl_skip: (kwarg only) Clients that support it, will
                avoid returning entries with ttls below this

        Returns: An iterator over the respective (key, value) pairs. When a
            miss occurs, the default is returned in every case (instead of
            an exception). It's up to the caller to check for a default
            return value. The order of iteration may not match the given
            iterator.
        """
        ttl_skip = kw.get('ttl_skip') or 0
        for key, (rv, ttl) in self.getTtlMulti(keys, default, **kw):
            if ttl < ttl_skip:
                yield key, default
            else:
                yield key, rv

    def getTtlMulti(self, keys, default = NONE, **kw):
        """
        Params:

            keys: An iterable of keys to fetch

            default: The default value to return when a miss occurs

            ttl_skip: (kwarg only) Clients that support it, will
                avoid returning entries with ttls below this

        Returns: An iterator over the respective (key, (value, ttl)) tuples.
            When a miss occurs, the default is returned in every case (instead
            of an exception). It's up to the caller to check for a default
            return value. The order of iteration may not match the given
            iterator.
        """
        for key in keys:
            try:
                yield key, self.getTtl(key, default, **kw)
            except CacheMissError:
                yield key, (default, -1)

    @abstractmethod
    def clear(self):
        raise NotImplementedError

    @abstractmethod
    def purge(self, timeout = 0):
        """
        Params

            timeout: if specified, only items that have been stale for
                this amount of time will be removed. That is, it is added to the
                initial entry's TTL vale.
        """
        raise NotImplementedError

    @abstractmethod
    def contains(self, key, ttl = None):
        """
        Verifies that a key is valid within the cache

        Params

            key: the key to check

            ttl: If provided and not None, a TTL margin. Keys with this or less
                time to live will be considered as stale. Provide if you want
                to check about-to-expire keys.
        """
        return False

class ReadWriteSyncAdapter(BaseCacheClient):
    def __init__(self, client):
        self.client = client

    @property
    def is_async(self):
        return self.client.is_async

    @property
    def capacity(self):
        return self.client.capacity

    @property
    def usage(self):
        return self.client.usage

    @serialize_write
    def put(self, key, value, ttl, **kw):
        return self.client.put(key, value, ttl, **kw)

    @serialize_write
    def renew(self, key, ttl, **kw):
        return self.client.renew(key, ttl, **kw)

    @serialize_write
    def add(self, key, value, ttl, **kw):
        return self.client.add(key, value, ttl, **kw)

    @serialize_write
    def delete(self, key):
        return self.client.delete(key)

    @serialize_write
    def expire(self, key):
        return self.client.expire(key)

    @serialize_read
    def getTtl(self, key, default = NONE, **kw):
        return self.client.getTtl(key, default, **kw)

    @serialize_read
    def get(self, key, default = NONE, **kw):
        return self.client.get(key, default, **kw)

    @serialize_read
    def getTtlMulti(self, keys, default = NONE, **kw):
        return self.client.getTtlMulti(keys, default, **kw)

    @serialize_read
    def getMulti(self, keys, default = NONE, **kw):
        return self.client.getMulti(keys, default, **kw)

    @serialize_write
    def promote(self, *p, **kw):
        return self.client.promote(*p, **kw)

    @serialize_write
    def clear(self):
        return self.client.clear()

    @serialize_write
    def purge(self, *p, **kw):
        return self.client.purge(*p, **kw)

    @serialize_read
    def contains(self, key, ttl = None, **kw):
        return self.client.contains(key, ttl, **kw)

    def __str__(self):
        return "<%s of %r>" % (self.__class__.__name__, self.client)

    def __repr__(self):
        return str(self)


class SyncAdapter(BaseCacheClient):
    def __init__(self, client):
        self.client = client

    @property
    def is_async(self):
        return self.client.is_async

    @property
    def capacity(self):
        return self.client.capacity

    @property
    def usage(self):
        return self.client.usage

    @serialize
    def put(self, key, value, ttl, **kw):
        return self.client.put(key, value, ttl, **kw)

    @serialize
    def renew(self, key, ttl, **kw):
        return self.client.renew(key, ttl, **kw)

    @serialize
    def add(self, key, value, ttl, **kw):
        return self.client.add(key, value, ttl, **kw)

    @serialize
    def delete(self, key):
        return self.client.delete(key)

    @serialize
    def expire(self, key):
        return self.client.expire(key)

    @serialize
    def getTtl(self, key, default = NONE, **kw):
        return self.client.getTtl(key, default, **kw)

    @serialize
    def get(self, key, default = NONE, **kw):
        return self.client.get(key, default, **kw)

    @serialize
    def getTtlMulti(self, keys, default = NONE, **kw):
        return self.client.getTtlMulti(keys, default, **kw)

    @serialize
    def getMulti(self, keys, default = NONE, **kw):
        return self.client.getMulti(keys, default, **kw)

    @serialize_write
    def promote(self, *p, **kw):
        return self.client.promote(*p, **kw)

    @serialize
    def clear(self):
        return self.client.clear()

    @serialize
    def purge(self, *p, **kw):
        return self.client.purge(*p, **kw)

    @serialize
    def contains(self, key, ttl = None, **kw):
        return self.client.contains(key, ttl, **kw)

    def __str__(self):
        return "<%s of %r>" % (self.__class__.__name__, self.client)

    def __repr__(self):
        return str(self)


class DecoratedWrapper(BaseCacheClient):
    """
    A namespace wrapper client will decorate keys with a namespace, making it possible
    to share one client among many sub-clients without key collisions.
    """
    def __init__(self, client,
            key_decorator = None, key_undecorator = None,
            value_decorator = None, value_undecorator = None):
        self.client = client
        self.key_decorator = key_decorator
        self.key_undecorator = key_undecorator
        self.value_decorator = value_decorator
        self.value_undecorator = value_undecorator

    @property
    def is_async(self):
        return self.client.is_async

    @property
    def capacity(self):
        return self.client.capacity

    @property
    def usage(self):
        return self.client.usage

    def wait(self, key, timeout = None):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        return self.client.wait(key, timeout)

    def put(self, key, value, ttl, **kw):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        value_decorator = self.value_decorator
        if value_decorator:
            if hasattr(value, 'undefer'):
                callable_ = value.callable_
                value.callable_ = lambda *p, **kw: value_decorator(callable_(*p, **kw))
            else:
                value = value_decorator(value)
        return self.client.put(key, value, ttl, **kw)

    def put_coherently(self, key, ttl, expired, future, callable_, **kw):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        value_decorator = self.value_decorator
        if value_decorator:
            callable_ = lambda *p, **kw: value_decorator(callable_(*p, **kw))
        return self.client.put_coherently(key, ttl, expired, future, callable_, **kw)

    def renew(self, key, ttl, **kw):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        return self.client.renew(key, ttl, **kw)

    def add(self, key, value, ttl, **kw):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        value_decorator = self.value_decorator
        if value_decorator:
            value = value_decorator(value)
        return self.client.add(key, value, ttl, **kw)

    def delete(self, key):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        return self.client.delete(key)

    def expire(self, key):
        key_decorator = self.key_decorator
        if key_decorator:
            key = key_decorator(key)
        return self.client.expire(key)

    def _wrapPromoteCallback(self, key_undecorator, kw):
        promote_callback = kw['promote_callback']
        def undecorating_callback(key, value, ttl):
            promote_callback(key_undecorator(key), value, ttl)
        kw['promote_callback'] = undecorating_callback

    def getTtl(self, key, default = NONE, **kw):
        key_decorator = self.key_decorator
        if key_decorator is not None:
            key = key_decorator(key)
            key_undecorator = self.key_undecorator
            if key_undecorator is not None and 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
        rv, ttl = self.client.getTtl(key, default, **kw)
        if rv is not default:
            value_undecorator = self.value_undecorator
            if value_undecorator is not None:
                rv = value_undecorator(rv)
        return rv, ttl

    def get(self, key, default = NONE, **kw):
        key_decorator = self.key_decorator
        if key_decorator is not None:
            key = key_decorator(key)
            key_undecorator = self.key_undecorator
            if key_undecorator is not None and 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
        rv = self.client.get(key, default, **kw)
        if rv is not default:
            value_undecorator = self.value_undecorator
            if value_undecorator is not None:
                rv = value_undecorator(rv)
        return rv

    def getTtlMulti(self, keys, default = NONE, **kw):
        key_decorator = self.key_decorator
        key_undecorator = self.key_undecorator
        if key_decorator is not None:
            if key_undecorator is not None and 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
            keys = map(key_decorator, keys)
        value_undecorator = self.value_undecorator
        for gkey, (rv, ttl) in self.client.getTtlMulti(keys, default, **kw):
            if rv is not default and value_undecorator is not None:
                rv = value_undecorator(rv)
            if key_undecorator is not None:
                gkey = key_undecorator(gkey)
            yield gkey, (rv, ttl)

    def getMulti(self, keys, default = NONE, **kw):
        key_decorator = self.key_decorator
        key_undecorator = self.key_undecorator
        if key_decorator is not None:
            if key_undecorator is not None and 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
            keys = map(key_decorator, keys)
        value_undecorator = self.value_undecorator
        for gkey, rv in self.client.getMulti(keys, default, **kw):
            if rv is not default and value_undecorator is not None:
                rv = value_undecorator(rv)
            if key_undecorator is not None:
                gkey = key_undecorator(gkey)
            yield gkey, rv

    def promote(self, key, *p, **kw):
        key_decorator = self.key_decorator
        if key_decorator is not None:
            key = key_decorator(key)
            key_undecorator = self.key_undecorator
            if key_undecorator is not None and 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
        self.client.promote(key, *p, **kw)

    def clear(self):
        return self.client.clear()

    def purge(self, *p, **kw):
        return self.client.purge(*p, **kw)

    def contains(self, key, ttl = None, **kw):
        key_decorator = self.key_decorator
        if key_decorator is not None:
            key = key_decorator(key)
        return self.client.contains(key, ttl, **kw)

class NamespaceWrapper(DecoratedWrapper):
    """
    A namespace wrapper client will decorate keys with a namespace, making it possible
    to share one client among many sub-clients without key collisions.
    """
    def __init__(self, namespace, client, revmark_ttl=FOREVER):
        super(NamespaceWrapper, self).__init__(client)
        self.namespace = namespace
        self.key_undecorator = operator.itemgetter(2)
        self.revmark_ttl = revmark_ttl

        try:
            self.revision = client.get((namespace,'REVMARK'), 0)
        except NoServersError:
            self.revision = 0

    key_decorator = property(
        operator.attrgetter('_key_decorator'),
        lambda self, value : None)

    def _key_decorator(self, key):
        return (self.namespace, self.revision, key)

    def getTtl(self, key, default = NONE, **kw):
        return self.client.getTtl(self._key_decorator(key), default, **kw)

    def get(self, key, default = NONE, **kw):
        return self.client.get(self._key_decorator(key), default, **kw)

    def getTtlMulti(self, keys, default = NONE, **kw):
        # Specialized implementation of DecoratedWrapper
        key_decorator = self._key_decorator
        key_undecorator = self.key_undecorator
        if key_decorator is not None:
            if 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
            keys = map(key_decorator, keys)
        for gkey, (rv, ttl) in self.client.getTtlMulti(keys, default, **kw):
            yield key_undecorator(gkey), (rv, ttl)

    def getMulti(self, keys, default = NONE, **kw):
        # Specialized implementation of DecoratedWrapper
        key_decorator = self._key_decorator
        key_undecorator = self.key_undecorator
        if key_decorator is not None:
            if 'promote_callback' in kw:
                self._wrapPromoteCallback(key_undecorator, kw)
            keys = map(key_decorator, keys)
        for gkey, rv in self.client.getMulti(keys, default, **kw):
            yield key_undecorator(gkey), rv

    def contains(self, key, ttl = None, **kw):
        return self.client.contains(self._key_decorator(key), ttl, **kw)

    def clear(self):
        # Cannot clear a shared client, so, instead, switch revisions
        self.revision += 1
        self.client.put((self.namespace, 'REVMARK'), self.revision, self.revmark_ttl)

    def __str__(self):
        return "<%s namespace %r/%r on %r>" % (self.__class__.__name__, self.namespace, self.revision, self.client)

    def __repr__(self):
        return str(self)

class NamespaceMirrorWrapper(NamespaceWrapper):
    """
    A namespace wrapper that takes its namespace info from the provided reference (mirrors it).
    It also takes the internal revision number, to inherit namespace changes created by cache clears.
    """
    def __init__(self, reference, client):
        super(NamespaceWrapper, self).__init__(client)
        self.reference = reference

    @property
    def revision(self):
        return self.reference.revision

    @revision.setter
    def revision(self, value):
        self.reference.revision = value

    @property
    def namespace(self):
        return self.reference.namespace

    @namespace.setter
    def namespace(self, value):  # lint:ok
        pass

    @property
    def key_undecorator(self):
        return self.reference.key_undecorator

    @key_undecorator.setter
    def key_undecorator(self, value):  # lint:ok
        pass

    @property
    def key_decorator(self):
        return self.reference.key_decorator

    @key_decorator.setter
    def key_decorator(self, value):  # lint:ok
        pass

    @property
    def revmark_ttl(self):
        return self.reference.revmark_ttl

    @revmark_ttl.setter
    def revmark_ttl(self, value):
        self.reference.revmark_ttl = value



# -*- coding: utf-8 -*-
import time
import threading
import weakref

from . import base

from clru.lrucache import LRUCache as Cache, IsThreadsafe as CacheIsThreadsafe

CacheMissError = base.CacheMissError

if not CacheIsThreadsafe:

    import warnings
    warnings.warn("LRUCache extension module not built in, "
        "using pure-python version which is not atomic and requires"
        "explicit synchronization. Decreased performance will be noticeable")
    del warnings

from clru.cuckoocache import (
    LazyCuckooCache as CuckooCache, CacheMissError as CuckooCacheMissError, IsThreadsafe as CuckooIsThreadsafe
)
assert issubclass(CuckooCacheMissError, CacheMissError)

if CacheIsThreadsafe and not CuckooIsThreadsafe:
    import warnings
    warnings.warn("CuckooCache extension module not built in, "
        "but LRUCache module was built. InprocCacheClient will be assumed "
        "thread-safe, you will need to wrap it in a synchronized adapter "
        "to be used with CuckooCache!")

_caches_mutex = threading.RLock()
_caches = weakref.WeakKeyDictionary()

def _register_inproc(cache):
    _caches[cache] = None

def cacheStats():
    global _caches
    global _caches_mutex

    with _caches_mutex:
        rv = {}
        for cache in _caches.keys():
            fname = cache.__name__

            # Sometimes, functions are different but named the same. Usually
            # they're related, so we aggregate those stats.
            ocsize, oclen = rv.get(fname, (0,0))
            rv[fname] = ( cache.store.size + ocsize, len(cache.store)+oclen )

        return rv

def cachePurge(timeout = 0, sleeptime = None):
    with _caches_mutex:
        caches = list(_caches.keys())

    for cache in caches:
        if sleeptime is not None:
            time.sleep(sleeptime)
        cache.purge(timeout)

def cacheClear():
    """
    Clear all @cached caches (use with moderation)

    >>> from chorde.decorators import cached
    >>> from chorde.clients.inproc import cacheClear, InprocCacheClient
    >>> import itertools
    >>> random = itertools.cycle(iter((0.9560342718892494, 0.9478274870593494, 2, 3, 4, 5, 6, 7, 8))).__next__
    >>> @cached(InprocCacheClient(1000), ttl = 6000)
    ... def f():
    ...     return random()
    ...
    >>> @cached(InprocCacheClient(1000), ttl = 6000)
    ... def g():
    ...     return random()
    ...
    >>> f()
    0.9560342718892494
    >>> g()
    0.9478274870593494
    >>> f()
    0.9560342718892494
    >>> g()
    0.9478274870593494
    >>> cacheClear()
    >>> f()
    2
    >>> g()
    3
    >>> f()
    2
    >>> g()
    3
    """

    with _caches_mutex:
        caches = list(_caches.keys())

    for cache in caches:
        cache.clear()

class CacheJanitorThread(threading.Thread):

    def __init__(self, sleep_interval, purge_timeout = 0):
        threading.Thread.__init__(self)
        self.sleep_interval = sleep_interval
        self.purge_timeout = purge_timeout
        self.logger = None
        self.setDaemon(True)

    def run(self):
        global cachePurge

        while True:
            time.sleep(self.sleep_interval)
            try:
                cachePurge(self.purge_timeout, sleeptime = 0.01)
            except:
                if self.logger is None:
                    pass
                else:
                    self.logger.error("Exception during cache purge", exc_info = True)

def startCacheJanitorThread(sleep_interval=3600, purge_timeout=0):
    thread = CacheJanitorThread(sleep_interval, purge_timeout)
    thread.start()
    return thread

class InprocCacheClient(base.BaseCacheClient):
    def __init__(self, size, store_class = None, store_args = (), store_kwargs = {}):
        if store_class is None:
            store_class = Cache
        self.store = store_class(size, *store_args, **store_kwargs)
        self.purgecount = 0
        _register_inproc(self)

    @property
    def is_async(self):
        return False

    @property
    def capacity(self):
        return self.store.size

    @property
    def usage(self):
        return len(self.store)

    def put(self, key, value, ttl, time = time.time):
        self.store[key] = (value, time() + ttl)

    def renew(self, key, ttl, baseNONE = base.NONE, time = time.time):
        origvalue = self.store.get(key, baseNONE)
        if origvalue is not baseNONE:
            value, kttl = origvalue
            ttl += time()
            if kttl < ttl:
                self.store.cas(key, origvalue, (value, ttl))

    def add(self, key, value, ttl):
        now = time.time()
        new_entry = (value, now + ttl)
        cur_entry = self.store.setdefault(key, new_entry)
        if cur_entry is new_entry:
            return True
        else:
            # Check TTL
            if cur_entry[1] < now:
                self.store[key] = new_entry
                return True
            else:
                return False

    def delete(self, key):
        try:
            del self.store[key]
        except CacheMissError:
            pass

    def expire(self, key):
        now = time.time()
        store = self.store
        try:
            value, ttl = store.pop(key)
            store.setdefault(key, (value, now))
        except CacheMissError:
            pass

    def getTtl(self, key, default = base.NONE, ttl_skip=None,
            promote_callback=None, baseNONE = base.NONE, time=time.time):
        rv = self.store.get(key, baseNONE)
        if rv is not baseNONE:
            rv, ttl = rv
            ttl = ttl - time()
            return rv, ttl
        elif default is baseNONE:
            raise CacheMissError(key)
        else:
            return default, -1

    def clear(self):
        self.store.clear()

    def purge(self, timeout = 0):
        deletions = []
        retentions = []
        cache = self.store
        curtime = time.time() - timeout
        try:
            deletions_append = deletions.append
            for k, (v, timestamp) in cache.iteritems():
                if timestamp < curtime:
                    deletions_append(k)
        except RuntimeError:
            # Store changed, try with a snapshot
            # Only needed if the store isn't the LRU implementation
            # which should only happen in really crippled systems
            del deletions[:]
            cache_get = cache.get
            deletions_append = deletions.append
            for k in cache.keys():
                v = cache_get(v)
                if v is not None:
                    v, timestamp = v
                    if timestamp < curtime:
                        deletions_append(k)
        retentions_append = retentions.append
        cache_pop = cache.pop
        for k in deletions:
            # keep them alive so that no finalizations occur within the mutex's scope
            # (when wrapped inside a ReadWriteSyncAdapter), otherwise weird deadlocks
            # could arise.
            retentions_append(cache_pop(k, None))
        self.purgecount += 1
        if self.purgecount > 16:
            self.store.defrag()
            self.purgecount = 0

        # Returning them makes them live at least until the sync-wrapped method ends
        return retentions

    def contains(self, key, ttl = None, baseNONE = base.NONE, time = time.time):
        rv = self.store.get(key, baseNONE)
        if rv is not baseNONE:
            if ttl is None:
                ttl = 0
            store_ttl = rv[1] - time()
            return store_ttl > ttl
        else:
            return False

if not CacheIsThreadsafe:
    InprocCacheClient_ = InprocCacheClient
    def InprocCacheClient(*p, **kw):
        return base.ReadWriteSyncAdapter(InprocCacheClient_(*p, **kw))




# -*- coding: utf-8 -*-
import time
import zmq
import logging
import json
import os
from collections import defaultdict
from past.builtins import unicode
from base64 import b64encode

import pickle
import io

__ALL__ = (
    'EVENT_INCOMING_UPDATE',
    'EVENT_UPDATE_ACKNOWLEDGED',
    'EVENT_UPDATE_IGNORED',
    'EVENT_UPDATE_SENT',
    'EVENT_ENTER_BROKER',
    'EVENT_LEAVE_BROKER',
    'EVENT_ENTER_LISTENER',
    'EVENT_LEAVE_LISTENER',
    'EVENT_IDLE',
    'EVENT_TIC',
    'EVENT_NAMES',

    'FRAME_UPDATE_OK',
    'FRAME_UPDATE_DROPPED',
    'FRAME_VALID_UPDATE_REPLIES',
    'EVENT_FOR_REPLY',

    'BrokerReply',
    'BaseIPSub',
)

FRAME_HEARTBEAT = "__HeyDude__"
FRAME_UPDATE_OK = "OK"
FRAME_UPDATE_DROPPED = "DROP"
FRAME_VALID_UPDATE_REPLIES = (FRAME_UPDATE_OK, FRAME_UPDATE_DROPPED)

# EVENT KEY                         message payload format
EVENT_INCOMING_UPDATE = 1         # [prefix, identity, payload]. Payload verbatim as sent
EVENT_UPDATE_ACKNOWLEDGED = 2     # (update, reply frames)
EVENT_UPDATE_IGNORED = 3          # (update, reply frames)
EVENT_UPDATE_SENT = 4             # [prefix, identity, payload]. Payload verbatim as sent
EVENT_ENTER_BROKER = 5            # None
EVENT_LEAVE_BROKER = 6            # None
EVENT_ENTER_LISTENER = 7          # None
EVENT_LEAVE_LISTENER = 8          # None
EVENT_IDLE = 9                    # None
EVENT_TIC = 10                    # None

EVENT_NAMES = {
    EVENT_INCOMING_UPDATE : 'INCOMING_UPDATE',
    EVENT_UPDATE_ACKNOWLEDGED : 'UPDATE_ACKNOWLEDGED',
    EVENT_UPDATE_IGNORED: 'UPDATE_DROPPED',
    EVENT_UPDATE_SENT : 'UPDATE_SENT',
    EVENT_ENTER_BROKER : 'ENTER_BROKER',
    EVENT_LEAVE_BROKER : 'LEAVE_BROKER',
    EVENT_ENTER_LISTENER : 'ENTER_LISTENER',
    EVENT_LEAVE_LISTENER : 'LEAVE_LISTENER',
    EVENT_IDLE : 'IDLE',
    EVENT_TIC : 'TIC',
}

IDENTITY_EVENTS = (EVENT_INCOMING_UPDATE,)

FRAME_UPDATE_OK = "OK"
FRAME_UPDATE_DROPPED = "DROP"
FRAME_VALID_UPDATE_REPLIES = (FRAME_UPDATE_OK, FRAME_UPDATE_DROPPED)

EVENT_FOR_REPLY = {
    FRAME_UPDATE_OK : EVENT_UPDATE_ACKNOWLEDGED,
    FRAME_UPDATE_DROPPED : EVENT_UPDATE_IGNORED,
}

MAX_PREFIX = 256

IDLE_PERIOD = 0.5
TIC_PERIOD = 1.0

ENCODINGS = {
    'json' : lambda x : json.dumps(x, separators=(',',':')),
    'bytes' : lambda x : x,
    'utf8' : lambda x : x.encode('utf8') if isinstance(x, unicode) else x,
}

DECODINGS = {
    'json' : json.loads,
    'bytes' : lambda x : x,
    'utf8' : lambda x : x.decode('utf8'),
}

STREAMDECODINGS = {
    'json' : json.load,
    'bytes' : lambda x : x.read(),
    'utf8' : lambda x : x.read().decode('utf8'),
}

class BrokerReply(object):
    __slots__ = ('reply',)
    def __init__(self, *reply_parts):
        self.reply = reply_parts

class BaseIPSub(object):
    # Implementations are free to provide alternative versions as static/classmethods
    def __init__(self):
        self.logger = logging.getLogger('chorde.ipsub')

        self.listeners = defaultdict(lambda : defaultdict(set))

        self.last_idle = time.time()
        self.last_tic = time.time()

        self.identity = b"%x-%x-%s" % (
            os.getpid(),
            id(self),
            b64encode(os.urandom(8)).strip(b'\t =\n'),
        )

    def _idle(self):
        """
        Implementations SHOULD call this method when they've had no
        messages to process for a while heartbeat period (approximately).
        """
        if EVENT_IDLE in self.listeners:
            # Rate-limit idle events
            if time.time() >= (self.last_idle + IDLE_PERIOD):
                self._notify_all(EVENT_IDLE, None)
                self.last_idle = time.time()

        # Take the opportunity to check tic timestamp
        self._tic()

    def _tic(self):
        if EVENT_TIC in self.listeners:
            # Rate-limit tic events
            if time.time() >= (self.last_tic + TIC_PERIOD):
                self._notify_all(EVENT_TIC, None)
                self.last_tic = time.time()

    def request_tic(self):
        """
        Request that an EVENT_TIC be issued at the next oportunity
        """
        try:
            self.last_tic = time.time() - TIC_PERIOD
            self._pushsocket().send("tic")
        except zmq.ZMQError:
            # Shit happens, probably not connected
            pass

    def listen(self, prefix, event, callback):
        """
        Registers a listener for all events whose prefix starts
        with the given prefix.

        The callback will be invoked with the whole prefix as
        first argument, or None if the event doesn't have one,
        the event id as second argument, and the whole message,
        including prefix and payload, as third argument.
        Use decode_payload to decode, if needed.

        It should return True, if it is to be called again, or
        False if the listener is to be removed. Designated
        brokers can also return a BrokerReply wrapper, in which case
        the reply's payload will be returned to the listener where
        the update originated, providing a way to piggy-back the
        req-response connection among them. These are considered
        as True, so they will not be automatically removed.

        Listeners are not guaranteed to be called in any specific
        or stable order, but they are guaranteed to be called just
        once (per instance, not function name). They should return fast,
        or the I/O thread may stall.
        """
        self.listeners[event][prefix].add(callback)
        if event in IDENTITY_EVENTS:
            # Those are external, so we must subscribe
            self.add_subscriptions((prefix,))

    def listen_decode(self, prefix, event, callback):
        """
        See listen. The difference is that in this case, the payload
        in decoded form will be given to the callback, rather than
        the entire message.

        Returns the actual callback to be used for unlistening.
        """
        def decoding_callback(prefix, event, message):
            return callback(prefix, event, self.decode_payload(message))
        self.listen(prefix, event, decoding_callback)
        return decoding_callback

    def unlisten(self, prefix, event, callback):
        """
        Removes the specified listener. See listen.
        """
        if prefix in self.listeners[event]:
            try:
                self.listeners[event][prefix].remove(callback)
            except KeyError:
                pass
        if event in IDENTITY_EVENTS and not self.listeners[event].get(prefix):
            # Not interesting anymore... unsubscribe
            self.cancel_subscriptions((prefix,))

    def _notify_all(self, event, update):
        """
        Notifies all listeners of the given event, with the payload being "update".
        See publish on the format of the payload.
        """
        listeners = self.listeners.get(event)
        if not listeners and self._ndebug:
            return

        if event in IDENTITY_EVENTS and len(update) > 1:
            identity = update[1]
        else:
            identity = None

        if self._ndebug is None:
            self._ndebug = not self.logger.isEnabledFor(logging.DEBUG)

        if not self._ndebug:
            if identity is None and event != EVENT_UPDATE_SENT:
                prefix = None
            elif len(update[0]) < MAX_PREFIX:
                prefix = update[0]
            else:
                prefix = update[0][:MAX_PREFIX]
            if identity is None or identity == self.identity:
                self.logger.debug("IPSub: (from %r) %s (prefix %r)", self.identity, EVENT_NAMES[event], prefix)
            else:
                self.logger.debug("IPSub: (from %r) %s (prefix %r)", identity, EVENT_NAMES[event], prefix)

        if identity is not None and identity == self.identity:
            # Ehm... identified roundtrip -> ignore
            return

        if listeners:
            if identity is None:
                prefix = None
            elif len(update[0]) < MAX_PREFIX:
                prefix = update[0]
            else:
                prefix = update[0][:MAX_PREFIX]
            called = set()
            rrv = rv = None
            for cb_prefix, callbacks in list(listeners.items()):
                if prefix is None or prefix.startswith(cb_prefix):
                    byebye = set()
                    for callback in set(callbacks):
                        if callback in called:
                            continue
                        try:
                            rv = callback(prefix, event, update)
                            if not rv:
                                byebye.add(callback)
                            else:
                                called.add(callback)
                                if isinstance(rv, BrokerReply):
                                    rrv = rv
                        except:
                            self.logger.error("Exception in handler", exc_info = True)
                            byebye.add(callback)
                    for callback in byebye:
                        self.unlisten(cb_prefix, event, callback)
            return rrv or rv

    def add_subscriptions(self, prefixes):
        """
        Mark the given set of prefixes as "interesting".
        If the underlying channel support filtering by message prefix, these
        prefixes that haven been added are the ones that must be captured.

        Implementations must provide this method.
        """
        raise NotImplementedError

    def cancel_subscriptions(self, prefixes):
        """
        Mark the given set of prefixes as "uninteresting".
        If the underlying channel support filtering by message prefix, these
        prefixes should not be captured any more.

        Implementations must provide this method.
        """
        raise NotImplementedError

    def publish_json(self, prefix, payload, timeout = None):
        """
        Publish the message using the 'json' encoding. See publish.
        """
        self.publish(prefix, ['json',ENCODINGS['json'](payload)], timeout)

    def publish_pyobj(self, prefix, payload, timeout = None):
        """
        Publish the message using the 'pyobj' encoding. See publish.
        """
        self.publish(prefix, ['pyobj',ENCODINGS['pyobj'](payload)], timeout)

    def publish_bytes(self, prefix, payload, timeout = None):
        """
        Publish the message using the 'bytes' encoding. See publish.
        """
        self.publish(prefix, ['bytes',ENCODINGS['bytes'](payload)], timeout)

    def publish_unicode(self, prefix, payload, timeout = None):
        """
        Publish the message using the 'utf8' encoding. See publish.
        """
        self.publish(prefix, ['utf8',ENCODINGS['utf8'](payload)], timeout)

    def publish_encode(self, prefix, encoding, payload, timeout = None):
        """
        Publish the message using a custom encoding (must be registered with register_encoding).
        See publish, register_encoding.
        """
        self.publish(prefix, self.encode_payload(encoding, payload), timeout)

    @staticmethod
    def register_encoding(name, encoder, decoder, stream_decoder):
        """
        Register an encoding with the specified name.

        Params:
            encoder: a callable that takes the object to be dumped,
                and returns a string or buffer object.
            decoder: a callable that takes a string resulting of encoder,
                and returns in the represented object
            stream_decoder: like decoder, but instead will take a
                file-like object.
        """
        ENCODINGS[name] = encoder
        DECODINGS[name] = decoder
        STREAMDECODINGS[name] = stream_decoder

    @staticmethod
    def register_pyobj(pickler, unpickler):
        """
        Registers a pickling encoding.

        Params:
            pickler, unpickler: pickler/unpickler factory callables
                that take a file-like object to dump into. Can be
                stdlib's Pickle/Unpickle classes, or cPickle's, or
                sPickles, they both work out-of-the-box.
        """
        def dumps(x):
            io = io.StringIO()
            p = pickler(io,2)
            p.dump(x)
            return io.getvalue()
        def loads(x):
            io = io.StringIO(x)
            p = unpickler(io)
            return p.load()
        def load(x):
            return unpickler(x).load()
        BaseIPSub.register_encoding('pyobj', dumps, loads, load)

    @staticmethod
    def register_default_pyobj():
        BaseIPSub.register_pyobj(pickle.Pickler, pickle.Unpickler)

    @staticmethod
    def encode_payload(encoding, payload):
        return [encoding, ENCODINGS[encoding](payload)]

    @classmethod
    def decode_payload(cls, payload):
        encoding = payload[-2]
        payload = payload[-1]
        return DECODINGS[encoding](payload)

    def publish(self, prefix, payload, copy = False, timeout = None):
        """
        Publish a message, with the given prefix and payload.

        If copy is True, a copy of the payload is made (especially useful if the payload comes
        from another message and contains reusable buffers instead of simple strings, for which
        copy=True would be a no-op).

        If timeout is given and not None, it specifies a maximum time to block in milliseconds
        while attempting to push the message. If the timeout expires an exception will be thrown.

        Implementations must provide this method.
        """
        raise NotImplementedError




# -*- coding: utf-8 -*-
import sys

from .inproc import InprocCacheClient # before base, really, it injects CacheMissError
from .base import BaseCacheClient, ReadWriteSyncAdapter, NONE, CacheMissError, TimeoutError, CancelledError
from .asyncache import AsyncWriteCacheClient, Defer, AsyncCacheProcessor
from .tiered import TieredInclusiveClient

__all__ = [
    "NONE",
    "CacheMissError",
    "TimeoutError",
    "CancelledError",
    "Defer",

    "AsyncCacheProcessor",
    "AsyncWriteCacheClient",
    "BaseCacheClient",
    "InprocCacheClient",
    "ReadWriteSyncAdapter",
    "TieredInclusiveClient",
]

# Optionals below
try:
    from .memcached import MemcachedClient, FastMemcachedClient  # lint:ok
    __all__.extend(["MemcachedClient", "FastMemcachedClient"])
except ImportError:
    pass

try:
    from .coherent import CoherentDefer, CoherentWrapperClient  # lint:ok
    __all__.extend(["CoherentWrapperClient", "CoherentDefer"])
except ImportError:
    pass

try:
    from .files import FilesCacheClient  # lint:ok
    __all__.extend(["FilesCacheClient"])
except:
    pass




# -*- coding: utf-8 -*-
from threading import Thread
import logging

_logger = logging.getLogger("WorkerThread")

_cleanupHooks = []

class NotDoneError(Exception):
    pass

def registerThreadCleanupFunction(f):
    _cleanupHooks.append(f)

def _callCleanupHooks():
    global _logger

    for f in _cleanupHooks:
        try:
            f()
        except:
            _logger.error("Exception ocurred within thread cleanup function:", exc_info = True)

class WorkerThread(Thread):
    def __init__(self,routine,*args,**kwargs):
        Thread.__init__(self)
        self.routine = routine
        self.args = args
        self.kwargs = kwargs
        self._rv = None
        self._exc = None
        self.force_threading = False
        self.running = False

    def run(self):
        global _logger, _callCleanupHooks
        try:
            try:
                self.running = True
                self._rv = self.routine(*self.args,**self.kwargs)
            finally:
                _callCleanupHooks()
                self.running = False
        except Exception as e:
            self._exc = e
            _logger.error("Exception ocurred in worker thread:", exc_info = True)

    def start(self):
        global _nothreads
        self.running = True
        Thread.start(self)

    def rv(self, nowait=False):
        if self.running:
            if nowait:
                raise NotDoneError()
            self.join()
        try:
            if self._exc is not None:
                raise self._exc
            return self._rv
        finally:
            self._exc = None
            self._rv = None



from .sPickle import SecurePickler, SecureUnpickler
from .clients import *
from .clients.inproc import Cache as LRUCache, CuckooCache as CuckooCache
from ._version import __version__

try:
    from .mq import *
except ImportError:
    import warnings
    warnings.warn("No messaging built in, some features may not work")
    del warnings





# -*- coding: utf-8 -*-
# This monster makes it all compatible with up to py2.0 :-o

__all__ = [
    "SecurePickler",
    "SecureUnpickler",
]

#lint:disable
try:
    from hashlib import sha256 as checksum_algo
except ImportError:
    try:
        from hashlib import sha1 as checksum_algo
    except ImportError:
        try:
            from hashlib import md5 as checksum_algo
        except ImportError:
            try:
                from sha import sha as checksum_algo
            except ImportError:
                from md5 import md5 as checksum_algo
#lint:enable
checksum_algo_name = checksum_algo.__name__.replace('openssl_','')

import hmac
import struct
import threading
from binascii import hexlify, unhexlify

import pickle as cPickle  # lint:ok
from io import BytesIO  # lint:ok

class SecurePickler(object):
    def __init__(self, checksum_key, file, *p, **kw):
        self.file = file
        self.checksum_key = checksum_key

        self.backing_class = kw.pop('backing_class', cPickle.Pickler)
        self.backing_args = (p, kw)
        self.local = threading.local()

    @property
    def buf(self):
        try:
            return self.local.buf
        except AttributeError:
            self.local.buf = buf = BytesIO()
            return buf

    @buf.deleter
    def buf(self):  # lint:ok
        del self.local.buf

    @property
    def pickler(self):
        try:
            return self.local.pickler
        except AttributeError:
            p, kw = self.backing_args
            self.local.pickler = pickler = self.backing_class(self.buf, *p, **kw)
            return pickler

    @pickler.deleter
    def pickler(self):  # lint:ok
        del self.local.pickler

    @property
    def persistent_id(self):
        return self.pickler.persistent_id

    @persistent_id.setter
    def persistent_id(self, value):  # lint:ok
        self.pickler.persistent_id = value

    def dump(self,val):
        # dump to underlying pickler, then pick up the results
        self.pickler.dump(val)
        rv = self.buf.getvalue()
        self.buf.seek(0)
        self.buf.truncate()

        # compute HMAC, and prepend to output
        md = hmac.HMAC(self.checksum_key, rv, checksum_algo).hexdigest().encode("ascii")
        self.file.write(hexlify(struct.pack('<L',len(rv))))
        self.file.write(md)
        self.file.write(rv)

class SecureUnpickler(object):
    def __init__(self, checksum_key, file, *p, **kw):
        self.file = file
        self.checksum_key = checksum_key

        self.backing_class = kw.pop('backing_class', cPickle.Unpickler)
        self.backing_args = (p, kw)
        self.local = threading.local()

    @property
    def buf(self):
        try:
            return self.local.buf
        except AttributeError:
            self.local.buf = buf = BytesIO()
            return buf

    @buf.deleter
    def buf(self):  # lint:ok
        del self.local.buf

    @property
    def unpickler(self):
        try:
            return self.local.unpickler
        except AttributeError:
            p, kw = self.backing_args
            self.local.unpickler = unpickler = self.backing_class(self.buf, *p, **kw)
            return unpickler

    @unpickler.deleter
    def unpickler(self):  # lint:ok
        del self.local.unpickler

    @property
    def persistent_load(self):
        return self.unpickler.persistent_load

    @persistent_load.setter
    def persistent_load(self, value):  # lint:ok
        self.unpickler.persistent_load = value

    def load(self, headlen = len(struct.pack('<L',0)) * 2):
        datalen = self.file.read(headlen)
        if not datalen:
            raise EOFError("Cannot read secure packet header")
        datalen, = struct.unpack('<L', unhexlify(datalen) )

        ref_md = hmac.HMAC(self.checksum_key, None, checksum_algo)
        md = self.file.read(ref_md.digest_size*2)

        data = self.file.read(datalen)
        ref_md.update(data)

        ref_md = ref_md.hexdigest().encode("ascii")
        if ref_md != md:
            raise ValueError("MAC mismatch unpickling")

        buf = self.buf
        buf.seek(0)
        buf.write(data)
        buf.truncate()
        buf.seek(0)
        rv = self.unpickler.load()
        buf.seek(0)
        buf.truncate()
        return rv

def dump(key, obj, file, *p, **kw):
    pickler = SecurePickler(key, file, *p, **kw)
    pickler.dump(obj)

def dumps(key, obj, *p, **kw):
    buf = BytesIO()
    pickler = SecurePickler(key, buf, *p, **kw)
    pickler.dump(obj)
    return buf.getvalue()

def load(key, file, *p, **kw):
    unpickler = SecureUnpickler(key, file, *p, **kw)
    return unpickler.load()

def loads(key, str, *p, **kw):
    buf = BytesIO(str)
    unpickler = SecureUnpickler(key, buf, *p, **kw)
    return unpickler.load()



# -*- coding: utf-8 -*-
import memcache
import socket
import six

from . import memcached

class ElastiCacheStoreClient(memcached.MemcachedStoreClient):
    """
    This is a low-level memcache client for use with ElastiCache, that can be passed
    to chorde.clients.memcached.MemcachedClient instead of the default memcache.Client,
    or simply use get_cluster_from_config_entrypoint to set up a regular MemcachedClient

    Tested to work with python-memcache 1.31+
    """

    @classmethod
    def get_cluster_from_config_entrypoint(cls, entrypoint, fallback = None):
        c = cls([entrypoint])
        cluster_description = c.get("AmazonElastiCache:cluster")
        if not cluster_description:
            cluster_description = c.get_config("cluster")
        if cluster_description:
            cluster_description = cluster_description.decode("utf8").split()
        if not cluster_description or len(cluster_description) <= 1:
            # Default, go directly to the configuration entry point
            return [entrypoint] if fallback is None else fallback
        else:
            cluster_nodes = cluster_description[1:]
            return [
                "%s:%d" % (host.rstrip('.'), int(port))
                for host, ip, port in [ node.split('|',2) for node in cluster_nodes ]
            ]

    def get_config(self, key):
        if isinstance(key, six.text_type):
            key = key.encode("utf8")
        self.check_key(key)
        server, key = self._get_server(key)
        if not server:
            return None

        def _unsafe_get():
            self._statlog("config get")

            try:
                server.send_cmd(b"config get %s" % (key,))
                rkey = flags = rlen = None

                rkey, flags, rlen, = self._expectconfigvalue(server, raise_exception=True)

                if not rkey:
                    return None
                try:
                    value = self._recv_value(server, flags, rlen)
                finally:
                    server.expect(b"END", raise_exception=True)
            except (memcache._Error, socket.error) as msg:
                if isinstance(msg, tuple): msg = msg[1]
                server.mark_dead(msg)
                return None

            return value

        try:
            return _unsafe_get()
        except memcache._ConnectionDeadError:
            # retry once
            try:
                if server.connect():
                    return _unsafe_get()
                return None
            except (memcache._ConnectionDeadError, socket.error) as msg:
                server.mark_dead(msg)
            return None

    def _expectconfigvalue(self, server, line=None, raise_exception=False):
        if not line:
            line = server.readline(raise_exception)

        if line and line[:6] == b"CONFIG":
            resp, rkey, flags, len = line.split()
            flags = int(flags)
            rlen = int(len)
            return (rkey, flags, rlen)
        else:
            return (None, None, None)

class ElastiCacheClientMixin:

    FORCE_IS_DYNAMIC = True

    def expand_entry(self, entry):
        last_expansions = getattr(self, '_last_entry_expansions', None)
        if last_expansions is None:
            self._last_entry_expansions = last_expansions = {}
        get_cluster = getattr(self._client_class, 'get_cluster_from_config_entrypoint',
            ElastiCacheStoreClient.get_cluster_from_config_entrypoint)
        try:
            expanded = last_expansions[entry] = get_cluster(entry, fallback = last_expansions.get(entry))
        except:
            expanded = last_expansions.get(entry, (entry,))
        return expanded

class ElastiCacheClient(ElastiCacheClientMixin, memcached.MemcachedClient):
    pass

class FastElastiCacheClient(ElastiCacheClientMixin, memcached.FastMemcachedClient):
    pass



# -*- coding: utf-8 -*-
import time
import logging

from .asyncache import Defer, REGET, CancelledError
from .base import BaseCacheClient, NONE

from chorde.mq import coherence

class CoherentDefer(Defer):
    """
    Wrap a callable in this, and pass it as a value to an AsyncWriteCacheClient,
    and the evaluation of the given callable will happen asynchronously. The cache
    will return stale entries until the callable is finished computing a new value.
    Computation will be aborted if the coherence protocol shows someone else already
    computing, or if the shared cache is re-validated somehow when this method is
    called.
    """

    def __init__(self, callable_, *args, **kwargs):
        """
        Params
            callable_, args, kwargs: see Defer. All others must be passed by keyword.

            manager: A CoherenceManager instance that coordinates the corresponding cache clients.

            expired: A callable of the sort CoherenceManager takes, to re-check expiration status
                against the shared cache.

            expire_private: An optional callable that will be called with the given key (see below)
                right before receiving an OOB update. It should mark any privately held value for
                that key as expired, since a fresher one will be coming OOB. Usually the expire
                method of the private cache client works for this purpose.

            key: The key associated with this computation, to be provided to the CoherenceManager.

            timeout: Coherence protocol timeout, in ms, if the peers don't answer in this time,
                detached operation will be assumed and computation will proceed.
                Default: whatever the heartbeat timeout is on the underlying IPSub channel

            wait_time: Whether and how long to wait for other node's computation to be done.
                Normally, the default value of 0, which means "no wait", is preferrable so as not
                to stall deferred computation workers. However, in quick computations, it may be
                beneficial to provide a small wait time, to decrease latency in case some node
                goes down. This deferred would then take it upon him to start computing, and the
                whole group could be spared a whole cycle (versus just waiting for the value to
                be needed again), trading thoughput for latency. A value of None will cause
                infinite waits.
        """
        self.manager = kwargs.pop('manager')
        self.expired = kwargs.pop('expired')
        self.expire_private = kwargs.pop('expire_private', None)
        self.fetch = kwargs.pop('fetch', None)
        self.key = kwargs.pop('key')
        self.timeout = kwargs.pop('timeout', None)
        if self.timeout is None:
            self.timeout = self.manager.ipsub.heartbeat_push_timeout
        self.wait_time = kwargs.pop('wait_time', 0)
        self.computed = False
        self.aborted = False
        super(CoherentDefer, self).__init__(callable_, *args, **kwargs)

    def undefer(self):
        logger = logging.getLogger('chorde.coherence')
        while True:
            if not self.expired():
                logger.debug("Not computing because already fresh, key=%r", self.key)
                if hasattr(self, 'future'):
                    if self.expire_private is not None:
                        self.expire_private(self.key)
                    return REGET
                else:
                    return NONE
            else:
                computer = self.manager.query_pending(self.key, self.expired, self.timeout, True)
                if computer is None:
                    # My turn
                    logger.debug("Acquired computer duty on key=%r", self.key)
                    try:
                        rv = super(CoherentDefer, self).undefer()
                        if rv is not NONE:
                            logger.debug("Computing done on key=%r", self.key)
                            self.computed = True
                        else:
                            logger.debug("Computing skipped on key=%r", self.key)
                            self.aborted = True
                    except:
                        logger.debug("Computing aborted on key=%r due to exception", self.key, exc_info=True)
                        self.aborted = True
                        raise
                    return rv
                elif computer is coherence.OOB_UPDATE and not self.expired():
                    # Skip, tiered caches will read it from the shared cache and push it downstream
                    logger.debug("Fresh value available OOB on key=%r", self.key)
                    if hasattr(self, 'future'):
                        if self.expire_private is not None:
                            self.expire_private(self.key)
                        return REGET
                    else:
                        return NONE
                elif self.wait_time != 0:
                    logger.debug("Waiting for computation on node %r on key=%r", computer, self.key)
                    if self.manager.wait_done(self.key, timeout = self.wait_time):
                        logger.debug("Computation done (was waiting on node %r) on key=%r", computer, self.key)
                        if hasattr(self, 'future'):
                            if self.expire_private is not None:
                                self.expire_private(self.key)
                            return REGET
                        else:
                            return NONE
                    else:
                        # retry
                        logger.debug("Computation still in progress (waiting on node %r) on key=%r", computer, self.key)
                        continue
                else:
                    logger.debug("Computation in progress on node %r key=%r, not waiting", computer, self.key)
                    if hasattr(self, 'future'):
                        # Must cancel it if we're not going to wait
                        self.future.cancel()
                        self.future.set_exception(CancelledError())
                    return NONE

    def done(self):
        if self.computed:
            self.manager.mark_done(self.key)
        elif self.aborted:
            self.manager.mark_aborted(self.key)
        super(CoherentDefer, self).done()

class CoherentWrapperClient(BaseCacheClient):
    """
    Client wrapper that publishes cache activity with the coherence protocol.
    Given a manager, it will invoke its methods to make sure other instances
    on that manager are notified of invalidations.

    It adds another method of putting, put_coherently, that will additionally
    ensure that only one node is working on computing the result (thust taking
    a callable rather than a value, as defers do).

    The regular put method will publish being done, if the manager is configured
    in quick refresh mode, but will not attempt to obtain a computation lock,
    resulting in less overhead, decent consistency, but some duplication of
    effort. Therefore, put_coherently should be applied to expensive computations.

    If the underlying client isn't asynchronous, put_coherently will implicitly
    undefer the values, executing the coherence protocol in the calling thread.
    """

    def __init__(self, client, manager, timeout = 2000):
        self.client = client
        self.manager = manager
        self.timeout = timeout

    @property
    def is_async(self):
        return self.client.is_async

    @property
    def capacity(self):
        return self.client.capacity

    @property
    def usage(self):
        return self.client.usage

    def wait(self, key, timeout = None):
        # Hey, look at that. Since locally it it all happens on a Defer,
        # we can just wait on the wrapped client first to wait for ourselves
        if timeout is not None:
            deadline = time.time() + timeout
        else:
            deadline = None
        self.client.wait(key, timeout)

        # But in the end we'll have to ask the manager to talk to the other nodes
        if deadline is not None:
            timeout = int(max(0, deadline - time().time()) * 1000)
        else:
            timeout = None
        self.manager.wait_done(key, timeout=timeout)

    def put(self, key, value, ttl, coherence_timeout = None, **kw):
        manager = self.manager
        if manager.quick_refresh:
            # In quick refresh mode, we publish all puts
            if self.is_async and isinstance(value, Defer) and not isinstance(value, CoherentDefer):
                callable_ = value.callable_
                def done_after(*p, **kw):
                    rv = callable_(*p, **kw)
                    if rv is not NONE:
                        manager.fire_done([key])
                    return rv
                value.callable_ = done_after
                self.client.put(key, value, ttl, **kw)
            else:
                self.client.put(key, value, ttl, **kw)
                manager.fire_done([key], timeout = coherence_timeout)
        else:
            self.client.put(key, value, ttl, **kw)

    def put_coherently(self, key, ttl, expired, future, callable_, *args, **kwargs):
        """
        Another method of putting, that will additionally ensure that only one node
        is working on computing the result. As such, it takes  a callable rather
        than a value, as defers do.

        If the underlying client isn't asynchronous, put_coherently will implicitly
        undefer the value, executing the coherence protocol in the calling thread.

        Params
            key, ttl: See put
            expired: A callable that will re-check expiration status of the key on
                a shared cache. If this function returns False at mid-execution of
                the coherence protocol, the protocol and the computation will be
                aborted (assuming the underlying client will now instead fetch
                values from the shared cache).
            future: (optional) A future to be associated with the deferred
                computation.
            callable_, args, kwargs: See Defer. In contrast to normal puts, the
                callable may not be invoked if some other node has the computation
                lock.
        """
        wait_time = kwargs.pop('wait_time', 0)
        put_kwargs = kwargs.pop('put_kwargs', None) or None
        value = CoherentDefer(
            callable_,
            key = key,
            manager = self.manager,
            expired = expired,
            timeout = self.timeout,
            wait_time = wait_time,
            *args, **kwargs )
        if future is not None:
            value.future = future
        deferred = None
        try:
            if not self.is_async:
                # cannot put Defer s, so undefer right now
                deferred = value
                value = value.undefer()
                if value is NONE:
                    # Abort
                    return
            self.client.put(key, value, ttl, **(put_kwargs or {}))
        finally:
            if deferred is not None:
                deferred.done()

    def renew(self, key, ttl):
        self.client.renew(key, ttl)

    def delete(self, key, coherence_timeout = None):
        self.client.delete(key)

        # Warn others
        self.manager.fire_deletion(key, timeout = coherence_timeout)

    def expire(self, key):
        self.client.expire(key)

    def clear(self, coherence_timeout = None):
        self.client.clear()

        # Warn others
        self.manager.fire_deletion(coherence.CLEAR, timeout = coherence_timeout)

    def purge(self, *p, **kw):
        self.client.purge(*p, **kw)

    def getTtl(self, key, default = NONE, **kw):
        return self.client.getTtl(key, default, **kw)

    def get(self, key, default = NONE, **kw):
        return self.client.get(key, default, **kw)

    def getTtlMulti(self, keys, default = NONE, **kw):
        return self.client.getTtlMulti(keys, default, **kw)

    def getMulti(self, keys, default = NONE, **kw):
        return self.client.getMulti(keys, default, **kw)

    def contains(self, key, ttl = None, **kw):
        return self.client.contains(key, ttl, **kw)

    def promote(self, key, *p, **kw):
        return self.client.promote(key, *p, **kw)



cimport cython

@cython.no_gc
cdef class CacheStats(object):
    cdef public long long hits
    cdef public long long misses
    cdef public long long errors
    cdef public long long sync_misses
    cdef public long long sync_errors
    cdef public double min_miss_time
    cdef public double max_miss_time
    cdef public double sum_miss_time
    cdef public double sum_miss_time_sq
    cdef public double wait_time
    cdef public list miss_time_histogram
    cdef public double miss_time_histogram_max
    cdef public long miss_time_histogram_bins

    @cython.locals(hbins=long)
    cpdef add_histo(self, double time)



import sys
import os.path

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

no_pyrex = False
try:
    from Cython.Distutils import build_ext, Extension
    from Cython.Build import cythonize
except:
    no_pyrex = True

VERSION = "1.0.0-a1"

version_path = os.path.join(os.path.dirname(__file__), 'lib', 'chorde', '_version.py')
if not os.path.exists(version_path):
    with open(version_path, "w") as version_file:
        pass
with open(version_path, "r+") as version_file:
    version_content = "__version__ = %r" % (VERSION,)
    if version_file.read() != version_content:
        version_file.seek(0)
        version_file.write(version_content)
        version_file.flush()
        version_file.truncate()

with open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding="utf8") as readme_file:
    readme = readme_file.read()

import re
import os.path
_extras_requirements = [
    ( re.compile(r'pyzmq'), ['mq'] ),
    ( re.compile(r'numpy'), ['shmem'] ),
    ( re.compile(r'memcache'), ['memcache','elasticache'] ),
    ( re.compile(r'dnspython'), ['memcache','elasticache'] ),
]
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), encoding="utf8") as requirements_file:
    all_requirements = list(filter(bool, [ r.strip() for r in requirements_file ]))
# Compute extras_requires and main requirements
requirements = []
extras_requirements = {}
for req in all_requirements:
    isbase = True
    for pat, extnames in _extras_requirements:
        if pat.match(req):
            isbase = False
            for extname in extnames:
                extras_requirements.setdefault(extname, []).append(req)
    if isbase:
        requirements.append(req)

extra = {}

packages = [
    "chorde",
    "chorde.clients",
    "chorde.mq",
    "chorde.mq.ipsub",
]

if not no_pyrex:

    if sys.platform.startswith('linux'):
        DEFAULT_COMPILE_ARGS = '-mtune=native'
    else:
        DEFAULT_COMPILE_ARGS = ''

    EXTRA_COMPILE_ARGS = os.environ.get('CXXFLAGS', os.environ.get('CFLAGS', DEFAULT_COMPILE_ARGS)).split()

    basedir = os.path.dirname(__file__)
    libdir = os.path.join(basedir, 'lib')

    ext_modules = cythonize(
        [
            Extension("chorde.clients._async", ["lib/chorde/clients/_async.pyx"],
                depends = ["lib/chorde/clients/_async.pxd"],
                cython_include_dirs = [libdir, os.path.join(libdir, "chorde", "clients")],
                extra_compile_args = [ "-O3" ] ),
            Extension("chorde.decorators", ["lib/chorde/decorators.py"]),
            Extension("chorde.clients.base", ["lib/chorde/clients/base.py"]),
            Extension("chorde.clients.inproc", ["lib/chorde/clients/inproc.py"]),
            Extension("chorde.clients.tiered", ["lib/chorde/clients/tiered.py"]),
            Extension("chorde.clients.asyncache", ["lib/chorde/clients/asyncache.py"]),
        ],
        include_path = [ libdir ]
    )

    for ext_module in ext_modules:
        ext_module.extra_compile_args.extend(EXTRA_COMPILE_ARGS)

    extra.update(dict(
        ext_modules = ext_modules,
        cmdclass = {'build_ext': build_ext},
        package_data = {
            'chorde.clients' : ['_async.pxd'],
        },
    ))

setup(
    name = "chorde",
    version = VERSION,
    description = "Clustered Caching Library",
    author = "Claudio Freire",
    author_email = "klaussfreire@gmail.com",
    url = "https://github.com/klaussfreire/chorde",
    license = "LGPLv3",
    long_description = readme,
    packages = packages,
    package_dir = {'':'lib'},

    tests_require = 'nose',
    test_suite = 'tests',

    install_requires = requirements,
    extras_require = extras_requirements,

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
        "Operating System :: OS Independent",
    ],
    zip_safe = False,
    **extra
)

